from typing import Any, Callable, Dict, Optional, Type, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider

from ...base import (
    DEFAULT_FORMAT_FUNC,
    BytesValue,
    DynamicTemplate,
    FileMixin,
    StringValue,
)
from ...constants import DEFAULT_IMAGE_MAX_NB_CHARS
from ...helpers import load_class_from_path
from ...registry import FILE_REGISTRY
from ...storages.base import BaseStorage
from ...storages.filesystem import FileSystemStorage
from ..base.image_generator import BaseImageGenerator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "ImageMixin",
    "DEFAULT_IMAGE_GENERATOR",
    "IMAGEKIT_IMAGE_GENERATOR",
    "PIL_IMAGE_GENERATOR",
    "WEASYPRINT_IMAGE_GENERATOR",
)

IMAGEKIT_IMAGE_GENERATOR = (
    "faker_file.providers.image.imgkit_generator.ImgkitImageGenerator"
)
PIL_IMAGE_GENERATOR = (
    "faker_file.providers.image.pil_generator.PilImageGenerator"
)
WEASYPRINT_IMAGE_GENERATOR = (
    "faker_file.providers.image.weasyprint_generator.WeasyPrintImageGenerator"
)

DEFAULT_IMAGE_GENERATOR = IMAGEKIT_IMAGE_GENERATOR


class ImageMixin(FileMixin):
    """Image mixin."""

    @overload
    def _image_file(
        self: "ImageMixin",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def _image_file(
        self: "ImageMixin",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def _image_file(
        self: "ImageMixin",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate an image file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param max_nb_chars: Max number of chars for the content.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        if image_generator_cls is None:
            image_generator_cls = DEFAULT_IMAGE_GENERATOR

        if isinstance(image_generator_cls, str):
            image_generator_cls = load_class_from_path(image_generator_cls)

        if not image_generator_kwargs:
            image_generator_kwargs = {}

        image_generator_kwargs["content_specs"] = {
            "max_nb_chars": max_nb_chars,
            "wrap_chars_after": wrap_chars_after,
        }

        image_generator = image_generator_cls(
            generator=self.generator,
            **image_generator_kwargs,
        )
        data = {"content": "", "filename": filename, "storage": storage}
        if isinstance(content, DynamicTemplate):
            _content = content
        else:
            _content = self._generate_text_content(
                max_nb_chars=max_nb_chars,
                wrap_chars_after=wrap_chars_after,
                content=content,
                format_func=format_func,
            )
            data["content"] = _content

        _raw_content = image_generator.generate(
            content=_content,
            data=data,
            provider=self,
        )

        if raw:
            raw_content = BytesValue(_raw_content)
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, _raw_content)

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from faker import Faker
from faker_file.providers.augment_file_from_dir import (
    AugmentFileFromDirProvider,
)
from faker_file.providers.augment_file_from_dir.augmenters import (
    textaugment_augmenter,
)
from faker_file.providers.docx_file import DocxFileProvider
from faker_file.providers.eml_file import EmlFileProvider
from faker_file.providers.odt_file import OdtFileProvider
from faker_file.providers.txt_file import TxtFileProvider

FAKER = Faker()
FAKER.add_provider(DocxFileProvider)
FAKER.add_provider(TxtFileProvider)
FAKER.add_provider(EmlFileProvider)
FAKER.add_provider(OdtFileProvider)
FAKER.add_provider(AugmentFileFromDirProvider)

# Create files to test `augment_file_from_dir` with
FAKER.docx_file()
FAKER.eml_file()
FAKER.odt_file()
FAKER.txt_file()

# We assume that directory "/tmp/tmp/" exists and contains
# files of `DOCX`, `EML`, `EPUB`, `ODT`, `PDF`, `RTF` or `TXT`
# formats. Valid values for `action` are: "random_deletion",
# "random_insertion", "random_swap" and "synonym_replacement" (default).
augmented_file = FAKER.augment_file_from_dir(
    source_dir_path="/tmp/tmp/",
    text_augmenter_cls=textaugment_augmenter.EDATextaugmentAugmenter,
    text_augmenter_kwargs={
        "action": "synonym_replacement",
    },
)



from faker import Faker
from faker_file.providers.random_file_from_dir import RandomFileFromDirProvider
from faker_file.providers.txt_file import TxtFileProvider

FAKER = Faker()
FAKER.add_provider(TxtFileProvider)
FAKER.add_provider(RandomFileFromDirProvider)

# Create files to test `random_file_from_dir` with
for __i in range(5):
    FAKER.txt_file()

# We assume that directory "/tmp/tmp/" exists and contains files.
random_file = FAKER.random_file_from_dir(
    source_dir_path="/tmp/tmp/",
    prefix="zzz",
)



# Generated by Django 4.1.3 on 2022-12-04 21:22

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Upload",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, unique=True)),
                ("description", models.TextField(blank=True, null=True)),
                ("file", models.FileField(upload_to="")),
            ],
            options={
                "verbose_name": "Upload",
                "verbose_name_plural": "Upload",
            },
        ),
    ]



import os
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, Optional, Union, overload

from faker.providers import BaseProvider

from ..base import BytesValue, FileMixin, StringValue, returns_list
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage
from .helpers.inner import create_inner_txt_file

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("ZipFileProvider",)


class ZipFileProvider(BaseProvider, FileMixin):
    """ZIP file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.zip_file import ZipFileProvider

        FAKER = Faker()
        FAKER.add_provider(ZipFileProvider)

        file = FAKER.zip_file()

    Usage example with options:

    .. code-block:: python

        from faker_file.providers.helpers.inner import create_inner_docx_file

        file = FAKER.zip_file(
            prefix="zzz_archive_",
            options={
                "count": 5,
                "create_inner_file_func": create_inner_docx_file,
                "create_inner_file_args": {
                    "prefix": "zzz_docx_file_",
                    "max_nb_chars": 1_024,
                },
                "directory": "zzz",
            },
        )

    Usage example of nested ZIPs:

    .. code-block:: python

        from faker_file.providers.helpers.inner import create_inner_zip_file

        file = FAKER.zip_file(
            options={
                "create_inner_file_func": create_inner_zip_file,
                "create_inner_file_args": {
                    "options": {
                        "create_inner_file_func": create_inner_docx_file,
                    },
                },
            },
        )

    If you want to see, which files were included inside the ZIP, check
    the ``file.data["files"]``.
    """

    extension: str = "zip"

    @overload
    def zip_file(
        self: "ZipFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def zip_file(
        self: "ZipFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Union[BytesValue, StringValue]: ...

    def zip_file(
        self: "ZipFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a ZIP file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param options: Options (non-structured) for complex types, such as ZIP.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        data: Dict[str, Any] = {
            "inner": {},
            "files": [],
            "filename": filename,
            "storage": storage,
        }
        fs_storage = FileSystemStorage()

        # Specific
        if options:
            """
            A complex case. Could be initialized as follows:

            .. code-block:: python

                zip_file = FAKER.zip_file(
                    prefix="zzz_archive_",
                    options={
                        "count": 5,
                        "create_inner_file_func": create_inner_docx_file,
                        "create_inner_file_args": {
                            "prefix": "zzz_file_",
                            "max_nb_chars": 1_024,
                            "content": "{{date}}\r\n{{text}}\r\n{{name}}",
                        },
                        "directory": "zzz",
                    },
                )
            """
            _count = options.get("count", 5)
            _create_inner_file_func = options.get(
                "create_inner_file_func", create_inner_txt_file
            )
            _create_inner_file_args = options.get("create_inner_file_args", {})
            _dir_path = Path("")
            _directory = options.get("directory", "")

        else:
            # Defaults
            _count = 5
            _create_inner_file_func = create_inner_txt_file
            _create_inner_file_args = {}
            _dir_path = Path("")
            _directory = ""

        _zip_content = BytesIO()
        with zipfile.ZipFile(_zip_content, "w") as __fake_file:
            _kwargs = {"generator": self.generator}
            _kwargs.update(_create_inner_file_args)

            # If _create_inner_file_func returns a list of values
            if returns_list(_create_inner_file_func):
                _files = _create_inner_file_func(
                    storage=fs_storage,
                    **_kwargs,
                )
                for __file in _files:
                    data["inner"][str(__file)] = __file
                    __file_abs_path = fs_storage.abspath(__file)
                    __fake_file.write(
                        __file_abs_path,
                        arcname=Path(_directory) / Path(__file).name,
                    )
                    os.remove(__file_abs_path)  # Clean up temporary files
                    data["files"].append(Path(_directory) / Path(__file).name)

            # If _create_inner_file_func returns a single value
            else:
                for __i in range(_count):
                    __file = _create_inner_file_func(
                        storage=fs_storage,
                        **_kwargs,
                    )
                    data["inner"][str(__file)] = __file
                    __file_abs_path = fs_storage.abspath(__file)
                    __fake_file.write(
                        __file_abs_path,
                        arcname=Path(_directory) / Path(__file).name,
                    )
                    os.remove(__file_abs_path)  # Clean up temporary files
                    data["files"].append(Path(_directory) / Path(__file).name)

        if raw:
            raw_content = BytesValue(_zip_content.getvalue())
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, _zip_content.getvalue())

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from pathlib import Path
from typing import Union

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("BaseTextExtractor",)


class BaseTextExtractor:
    """Base text extractor."""

    path: str

    def __init__(
        self: "BaseTextExtractor",
        **kwargs,
    ) -> None:
        """Constructor.

        :param kwargs: Dictionary with parameters (for text extractor
            specific tuning).
        """
        self.handle_kwargs(**kwargs)

    def handle_kwargs(self: "BaseTextExtractor", **kwargs):
        """Handle kwargs."""

    def extract(
        self: "BaseTextExtractor",
        source_file: Union[Path, str],
    ) -> str:
        raise NotImplementedError("Method `extract` is not implemented.")



import os
import tempfile
from typing import Optional

from ..base import DEFAULT_REL_PATH
from .base import BaseStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("FileSystemStorage",)


class FileSystemStorage(BaseStorage):
    """File storage.

    Usage example:

    .. code-block:: python

        from faker_file.storages.filesystem import FileSystemStorage

        storage = FileSystemStorage()
        file = storage.generate_filename(prefix="zzz_", extension="docx")
        storage.write_text(file, "Lorem ipsum")
        storage.write_bytes(file, b"Lorem ipsum")

    Initialization with params:

    .. code-block:: python

        storage = FileSystemStorage()
    """

    def __init__(
        self: "FileSystemStorage",
        root_path: Optional[str] = tempfile.gettempdir(),
        rel_path: Optional[str] = DEFAULT_REL_PATH,
        *args,
        **kwargs,
    ) -> None:
        """
        :param root_path: Path of your files root directory (in case of Django
            it would be `settings.MEDIA_ROOT`).
        :param rel_path: Relative path (from root directory).
        """
        self.root_path = root_path or ""
        self.rel_path = rel_path or ""
        super().__init__(*args, **kwargs)

    def generate_filename(
        self: "FileSystemStorage",
        extension: str,
        prefix: Optional[str] = None,
        basename: Optional[str] = None,
    ) -> str:
        """Generate filename."""
        dir_path = os.path.join(self.root_path, self.rel_path)
        os.makedirs(dir_path, exist_ok=True)
        if not extension:
            raise Exception("Extension shall be given!")

        if not basename:
            basename = self.generate_basename(prefix)

        return os.path.join(dir_path, f"{basename}.{extension}")

        # if basename:
        #     return os.path.join(dir_path, f"{basename}.{extension}")
        # else:
        #     with tempfile.NamedTemporaryFile(
        #         prefix=prefix,
        #         dir=dir_path,
        #         suffix=f".{extension}",
        #     ) as temp_file:
        #         return temp_file.name

    def write_text(
        self: "FileSystemStorage",
        filename: str,
        data: str,
        encoding: Optional[str] = None,
    ) -> int:
        """Write text."""
        with open(filename, "w") as file:
            return file.write(data)

    def write_bytes(
        self: "FileSystemStorage",
        filename: str,
        data: bytes,
    ) -> int:
        """Write bytes."""
        with open(filename, "wb") as file:
            return file.write(data)

    def exists(self: "FileSystemStorage", filename: str) -> bool:
        """Check if file exists."""
        if os.path.isabs(filename):
            return os.path.exists(filename)
        return os.path.exists(os.path.join(self.root_path, filename))

    def relpath(self: "FileSystemStorage", filename: str) -> str:
        """Return relative path."""
        return os.path.relpath(filename, self.root_path)

    def abspath(self: "FileSystemStorage", filename: str) -> str:
        """Return absolute path."""
        if os.path.isabs(filename):
            return os.path.abspath(filename)
        return os.path.abspath(os.path.join(self.root_path, filename))

    def unlink(self: "FileSystemStorage", filename: str) -> None:
        """Delete the file."""
        abs_path = self.abspath(filename)
        os.remove(abs_path)



from django.urls import include, re_path
from rest_framework import routers

from .views import UploadViewSet

__all__ = ("urlpatterns",)

app_name = "api"
router = routers.DefaultRouter()
router.register(r"upload", UploadViewSet)

urlpatterns = [
    re_path("^upload/", include(router.urls)),
]



from faker import Faker
from faker_file.providers.mp3_file import Mp3FileProvider
from faker_file.providers.mp3_file.generators.gtts_generator import (
    GttsMp3Generator,
)

FAKER = Faker()
FAKER.add_provider(Mp3FileProvider)

mp3_file = FAKER.mp3_file(mp3_generator_cls=GttsMp3Generator)



from faker import Faker
from faker_file.base import DynamicTemplate
from faker_file.contrib.odt_file import (
    add_page_break,
    add_paragraph,
    add_picture,
    add_table,
)
from faker_file.providers.odt_file import OdtFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(OdtFileProvider)  # Register OdtFileProvider

# Create a ODT file with paragraph, picture, table and manual page breaks
# in between the mentioned elements. The ``DynamicTemplate`` simply
# accepts a list of callables (such as ``add_paragraph``,
# ``add_page_break``) and dictionary to be later on fed to the callables
# as keyword arguments for customising the default values.
odt_file = FAKER.odt_file(
    content=DynamicTemplate(
        [
            (add_paragraph, {}),  # Add paragraph
            (add_page_break, {}),  # Add page break
            (add_picture, {}),  # Add picture
            (add_page_break, {}),  # Add page break
            (add_table, {}),  # Add table
            (add_page_break, {}),  # Add page break
        ]
    )
)



from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, re_path
from drf_spectacular.views import (
    SpectacularAPIView,
    SpectacularRedocView,
    SpectacularSwaggerView,
)

__all__ = ("urlpatterns",)

urlpatterns = [
    re_path(r"^admin/", admin.site.urls),
    # API schema
    re_path("^api/schema/", SpectacularAPIView.as_view(), name="api-schema"),
    re_path(
        "^api/docs/",
        SpectacularSwaggerView.as_view(url_name="api-schema"),
        name="api-docs",
    ),
    re_path(
        "^api/redoc/",
        SpectacularRedocView.as_view(url_name="api-schema"),
        name="api-redoc",
    ),
    # Upload API
    re_path(r"^api/", include("upload.api.urls", namespace="api")),
]

# Serving media and static in debug/developer mode.
if settings.DEBUG:
    urlpatterns += staticfiles_urlpatterns()
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)



from multiprocessing import Pool

from faker import Faker
from faker_file.providers.helpers.inner import create_inner_docx_file
from faker_file.storages.filesystem import FileSystemStorage

FAKER = Faker()
STORAGE = FileSystemStorage()

# Document template
TEMPLATE = "Hey {{name}},\n{{text}},\nBest regards\n{{name}}"

with Pool(processes=2) as pool:
    for _ in range(10):  # Number of times we want to run our function
        pool.apply_async(
            create_inner_docx_file,
            # Apply async doesn't support kwargs. We have to pass all
            # arguments.
            [STORAGE, "mp", FAKER, None, None, TEMPLATE],
        )
    pool.close()
    pool.join()



# Required imports
from faker import Faker
from faker_file.providers.odt_file import OdtFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(OdtFileProvider)  # Register OdtFileProvider

# Generate ODT file
odt_file = FAKER.odt_file()



from faker import Faker
from faker_file.providers.pdf_file import PdfFileProvider
from faker_file.providers.pdf_file.generators.pil_generator import (
    PilPdfGenerator,
)

FAKER = Faker()
FAKER.add_provider(PdfFileProvider)

pdf_file = FAKER.pdf_file(pdf_generator_cls=PilPdfGenerator)



from faker import Faker
from faker_file.providers.docx_file import DocxFileProvider
from faker_file.providers.file_from_path import FileFromPathProvider

FAKER = Faker()
FAKER.add_provider(FileFromPathProvider)
FAKER.add_provider(DocxFileProvider)

# Create a file to use
docx_file = FAKER.docx_file(basename="file")

# We assume that directory "/tmp/tmp/" exists and contains a file named
# "file.docx".
docx_file_copy = FAKER.file_from_path(
    path="/tmp/tmp/file.docx",
    prefix="zzz",
)



import sys
import unittest
from typing import List, Union

from ..base import BytesValue, StringList, StringValue, returns_list

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("StringListTestCase",)


class StringListTestCase(unittest.TestCase):
    """StringList test case."""

    def test_string_list(self: "StringListTestCase") -> None:
        my_string = StringList(["apple", "banana", "cherry"], separator=" | ")

        self.assertEqual(str(my_string), "apple | banana | cherry")

        my_string.add_string("orange")
        self.assertEqual(str(my_string), "apple | banana | cherry | orange")

        my_string.remove_string("banana")
        self.assertEqual(str(my_string), "apple | cherry | orange")

        my_string += "grape"
        self.assertEqual(str(my_string), "apple | cherry | orange | grape")
        self.assertEqual(repr(my_string), "apple | cherry | orange | grape")


class ReturnsListTestCase(unittest.TestCase):
    def test_returns_list(self: "ReturnsListTestCase") -> None:
        def func_1() -> Union[BytesValue, StringValue]:
            """Returns Union[BytesValue, StringValue]"""
            return StringValue("func_1")

        def func_2() -> Union[StringValue, BytesValue]:
            """Returns Union[StringValue, BytesValue]"""
            return BytesValue(b"func_2")

        def func_3() -> List[Union[BytesValue, StringValue]]:
            """Returns List[Union[BytesValue, StringValue]]"""
            return [BytesValue(b"func_3"), StringValue("func_3")]

        def func_4() -> List[Union[StringValue, BytesValue]]:
            """Returns List[Union[StringValue, BytesValue]]"""
            return [StringValue("func_4"), BytesValue(b"func_4")]

        self.assertFalse(returns_list(func_1))
        self.assertFalse(returns_list(func_2))
        self.assertTrue(returns_list(func_3))
        self.assertTrue(returns_list(func_4))

    def test_no_return_hint(self: "ReturnsListTestCase"):
        def func_no_hint(a, b):
            pass

        self.assertFalse(returns_list(func_no_hint))

    def test_incorrect_hint(self: "ReturnsListTestCase"):
        def func_incorrect_hint(a) -> str:
            return "func_incorrect_hint"

        self.assertFalse(returns_list(func_incorrect_hint))

    def test_undefined_type_hint(self: "ReturnsListTestCase"):
        def func_undefined_hint(
            a: "UndefinedType",  # noqa
        ) -> List[Union[BytesValue, StringValue]]:
            return [BytesValue(b"fn_undef_hint"), StringValue("fn_undef_hint")]

        self.assertFalse(returns_list(func_undefined_hint))

    def test_correct_hint(self: "ReturnsListTestCase"):
        def func_correct_hint(a) -> List[Union[BytesValue, StringValue]]:
            return [BytesValue(b"fn_corr_hint"), StringValue("fn_corr_hint")]

        self.assertTrue(returns_list(func_correct_hint))

    def test_correct_hint_reversed(self: "ReturnsListTestCase"):
        def func_correct_hint_reversed(
            a,
        ) -> List[Union[StringValue, BytesValue]]:
            return [BytesValue(b"fn_corr_hint"), StringValue("fn_corr_hint")]

        self.assertTrue(returns_list(func_correct_hint_reversed))

    @unittest.skipIf(sys.version_info < (3, 9), "Skip on Python < 3.9")
    def test_correct_hint_builtin_list(self: "ReturnsListTestCase"):
        def func_correct_hint_builtin_list(
            a,
        ) -> list[Union[BytesValue, StringValue]]:
            return [BytesValue(b"fn_corr_hint"), StringValue("fn_corr_hint")]

        self.assertTrue(returns_list(func_correct_hint_builtin_list))

    @unittest.skipIf(sys.version_info < (3, 9), "Skip on Python < 3.9")
    def test_correct_hint_builtin_list_reversed(self: "ReturnsListTestCase"):
        def func_correct_hint_builtin_list_reversed(
            a,
        ) -> list[Union[StringValue, BytesValue]]:
            return [BytesValue(b"fn_corr_hint"), StringValue("fn_corr_hint")]

        self.assertTrue(returns_list(func_correct_hint_builtin_list_reversed))



"""
Created for testing purposes to mock out imaginary MaryTTS module
that is referenced in documentation examples.
"""


class MaryTTS:
    def __init__(self, locale: str, voice: str) -> None:
        self.locale = locale
        self.voice = voice

    def synth_mp3(self, content: str) -> bytes:
        return bytes()



import contextlib
import io
import logging
from typing import Any, Dict, Union

import imgkit
from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider

from ...base import DynamicTemplate, StringList
from ...constants import DEFAULT_FILE_ENCODING
from ..base.image_generator import BaseImageGenerator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("ImgkitImageGenerator",)

LOGGER = logging.getLogger(__name__)


class ImgkitImageGenerator(BaseImageGenerator):
    """Imgkit image generator.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.png_file import PngFileProvider
        from faker_file.providers.image.imgkit_generator import (
            ImgkitImageGenerator
        )

        FAKER = Faker()
        FAKER.add_provider(PngFileProvider)

        file = FAKER.png_file(
            img_generator_cls=ImgkitImageGenerator
        )


    Using `DynamicTemplate`:

    .. code-block:: python

        from faker_file.base import DynamicTemplate
        from faker_file.contrib.image.imgkit_snippets import (
            add_h1_heading,
            add_h2_heading,
            add_h3_heading,
            add_h4_heading,
            add_h5_heading,
            add_h6_heading,
            add_heading,
            add_page_break,
            add_paragraph,
            add_picture,
            add_table,
        )

        # Create a file with lots of elements
        file = FAKER.png_file(
            image_generator_cls=ImgkitImageGenerator,
            content=DynamicTemplate(
                [
                    (add_h1_heading, {}),
                    (add_paragraph, {}),
                    (add_h2_heading, {}),
                    (add_h3_heading, {}),
                    (add_h4_heading, {}),
                    (add_h5_heading, {}),
                    (add_h6_heading, {}),
                    (add_paragraph, {}),
                    (add_picture, {}),
                    (add_page_break, {}),
                    (add_h6_heading, {}),
                    (add_table, {}),
                    (add_paragraph, {}),
                ]
            )
        )
    """

    encoding: str = DEFAULT_FILE_ENCODING

    def handle_kwargs(self: "ImgkitImageGenerator", **kwargs) -> None:
        """Handle kwargs."""

    def generate(
        self: "ImgkitImageGenerator",
        content: str,
        data: Dict[str, Any],
        provider: Union[Faker, Generator, Provider],
        **kwargs,
    ) -> bytes:
        """Generate image."""
        with contextlib.redirect_stdout(io.StringIO()):
            if isinstance(content, DynamicTemplate):
                _content = StringList()
                for counter, (ct_modifier, ct_modifier_kwargs) in enumerate(
                    content.content_modifiers
                ):
                    ct_modifier(
                        provider,
                        self,
                        _content,
                        data,
                        counter,
                        **ct_modifier_kwargs,
                    )
            else:
                _content = (
                    f"<pre style='white-space: pre-wrap;'>{content}</pre>"
                )

            return imgkit.from_string(
                str(_content),
                False,
                options={"format": provider.extension},
            )



from faker import Faker
from faker_file.providers.docx_file import DocxFileProvider
from faker_file.providers.random_file_from_dir import RandomFileFromDirProvider

FAKER = Faker()
FAKER.add_provider(RandomFileFromDirProvider)
FAKER.add_provider(DocxFileProvider)

# Create files to use
for __i in range(10):
    FAKER.docx_file(basename="file")

# We assume that directory "/tmp/tmp/" exists and contains files with".docx"
# extension.
docx_file_copy = FAKER.random_file_from_dir(
    source_dir_path="/tmp/tmp/",
    prefix="zzz",
)



from typing import Any, Dict, Type

import nltk
from textaugment import EDA

from ...base.text_augmenter import BaseTextAugmenter

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "EDATextaugmentAugmenter",
    "DEFAULT_ACTION",
)

DEFAULT_ACTION = "synonym_replacement"


class EDATextaugmentAugmenter(BaseTextAugmenter):
    """Text extractor based on `ContextualWordEmbsAug` of `nlpaug`.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.augment_file_from_dir import (
            AugmentFileFromDirProvider,
        )
        from faker_file.providers.augment_file_from_dir.augmenters import (
            textaugment_augmenter,
        )

        FAKER = Faker()
        FAKER.add_provider(AugmentFileFromDirProvider)

        file = FAKER.augment_file_from_dir(
            source_dir_path="/tmp/tmp/",
            text_augmenter_cls=(
                textaugment_augmenter.EDATextaugmentAugmenter
            ),
            text_augmenter_kwargs={
                "action": "random_insertion",
            }
        )

    Refer to `textaugment` official documentation and check examples for `EDA`:

        ://github.com/dsfsi/textaugment/blob/master/examples/eda_example.ipynb

    Options for `action` are:

        - random_deletion
        - random_insertion
        - random_swap
        - synonym_replacement
    """

    action: str = DEFAULT_ACTION
    kwargs: Dict[str, Any] = {}
    _data_downloaded: bool = False

    def handle_kwargs(self: "EDATextaugmentAugmenter", **kwargs) -> None:
        """Handle kwargs."""
        if "action" in kwargs:
            self.action = kwargs["action"]
        if "kwargs" in kwargs:
            self.kwargs = kwargs["kwargs"]
        self.__class__.init()

    @classmethod
    def init(cls: Type["EDATextaugmentAugmenter"]) -> None:
        """Initialize the class. Should be run only once."""
        if not cls._data_downloaded:
            nltk.download("punkt")
            nltk.download("wordnet")
            nltk.download("stopwords")
            cls._data_downloaded = True

    def augment(self: "EDATextaugmentAugmenter", text: str) -> str:
        """Augment text."""
        aug = EDA()
        func = getattr(aug, self.action)
        return func(text, **self.kwargs)



from faker import Faker
from faker_file.providers.png_file import GraphicPngFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(GraphicPngFileProvider)  # Register provider

png_file = FAKER.graphic_png_file(size=(800, 800))



from faker import Faker
from faker_file.providers.image.imgkit_generator import ImgkitImageGenerator
from faker_file.providers.png_file import PngFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(PngFileProvider)  # Register PngFileProvider

# Generate PNG file using `imgkit`
pdf_file = FAKER.png_file(image_generator_cls=ImgkitImageGenerator)



import os
import tarfile
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, Optional, Union, overload

from faker.providers import BaseProvider

from ..base import BytesValue, FileMixin, StringValue, returns_list
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage
from .helpers.inner import create_inner_txt_file

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("TarFileProvider",)

COMPRESSION_OPTIONS = {"gz", "bz2", "xz"}


class TarFileProvider(BaseProvider, FileMixin):
    """TAR file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.tar_file import TarFileProvider

        FAKER = Faker()
        FAKER.add_provider(TarFileProvider)

        file = FAKER.tar_file()

    Usage example with options:

    .. code-block:: python

        from faker_file.providers.helpers.inner import create_inner_docx_file
        from faker_file.providers.tar_file import TarFileProvider

        file = FAKER.tar_file(
            prefix="ttt_archive_",
            options={
                "count": 5,
                "create_inner_file_func": create_inner_docx_file,
                "create_inner_file_args": {
                    "prefix": "ttt_docx_file_",
                    "max_nb_chars": 1_024,
                },
                "directory": "ttt",
            },
        )

    Usage example of nested TARs:

    .. code-block:: python

        from faker_file.providers.helpers.inner import create_inner_tar_file

        file = FAKER.tar_file(
            options={
                "create_inner_file_func": create_inner_tar_file,
                "create_inner_file_args": {
                    "options": {
                        "create_inner_file_func": create_inner_docx_file,
                    }
                }
            },
        )

    If you want to see, which files were included inside the TAR, check
    the ``file.data["files"]``.
    """

    extension: str = "tar"

    @overload
    def tar_file(
        self: "TarFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        # Once Python 3.7 is deprecated, add the following annotation:
        #     Optional[Literal["gz", "bz2", "xz"]] = None
        compression: Optional[str] = None,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def tar_file(
        self: "TarFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        # Once Python 3.7 is deprecated, add the following annotation:
        #     Optional[Literal["gz", "bz2", "xz"]] = None
        compression: Optional[str] = None,
        **kwargs,
    ) -> StringValue: ...

    def tar_file(
        self: "TarFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        # Once Python 3.7 is deprecated, add the following annotation:
        #     Optional[Literal["gz", "bz2", "xz"]] = None
        compression: Optional[str] = None,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a TAR file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param options: Options (non-structured) for complex types, such as ZIP.
        :param compression: Desired compression. Can be None or `gz`, `bz2`
            or `xz`.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.

        A complex case usage example:

        .. code-block:: python

            tar_file = TarFileProvider(None).tar_file(
                prefix="ttt_archive_",
                options={
                    "count": 5,
                    "create_inner_file_func": create_inner_docx_file,
                    "create_inner_file_args": {
                        "prefix": "ttt_file_",
                        "max_nb_chars": 1_024,
                        "content": "{{date}}\r\n{{text}}\r\n{{name}}",
                    },
                    "directory": "ttt",
                },
            )
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )
        data: Dict[str, Any] = {
            "inner": {},
            "files": [],
            "filename": filename,
            "storage": storage,
        }
        fs_storage = FileSystemStorage()

        # Specific
        if options:
            # Complex case
            _count = options.get("count", 5)
            _create_inner_file_func = options.get(
                "create_inner_file_func", create_inner_txt_file
            )
            _create_inner_file_args = options.get("create_inner_file_args", {})
            _dir_path = Path("")
            _directory = options.get("directory", "")

        else:
            # Defaults, simple case
            _count = 5
            _create_inner_file_func = create_inner_txt_file
            _create_inner_file_args = {}
            _dir_path = Path("")
            _directory = ""

        _tar_content = BytesIO()
        _mode = "w"
        if compression and compression in COMPRESSION_OPTIONS:
            _mode += f":{compression}"
        with tarfile.open(fileobj=_tar_content, mode=_mode) as __fake_file:
            _kwargs = {"generator": self.generator}
            _kwargs.update(_create_inner_file_args)

            # If _create_inner_file_func returns a list of values
            if returns_list(_create_inner_file_func):
                _files = _create_inner_file_func(
                    storage=fs_storage,
                    **_kwargs,
                )
                for __file in _files:
                    data["inner"][str(__file)] = __file
                    __file_abs_path = fs_storage.abspath(__file)
                    __fake_file.add(
                        __file_abs_path,
                        arcname=Path(_directory) / Path(__file).name,
                    )
                    os.remove(__file_abs_path)  # Clean up temporary files
                    data["files"].append(Path(_directory) / Path(__file).name)

            # If _create_inner_file_func returns a single value
            else:
                for __i in range(_count):
                    __file = _create_inner_file_func(
                        storage=fs_storage,
                        **_kwargs,
                    )
                    data["inner"][str(__file)] = __file
                    __file_abs_path = fs_storage.abspath(__file)
                    __fake_file.add(
                        __file_abs_path,
                        arcname=Path(_directory) / Path(__file).name,
                    )
                    os.remove(__file_abs_path)  # Clean up temporary files
                    data["files"].append(Path(_directory) / Path(__file).name)

        if raw:
            raw_content = BytesValue(_tar_content.getvalue())
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, _tar_content.getvalue())

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from faker import Faker
from faker_file.providers.pdf_file import PdfFileProvider

FAKER = Faker()
FAKER.add_provider(PdfFileProvider)

TEMPLATE = """
{{date}} {{city}}, {{country}}

Hello {{name}},

{{text}} {{text}} {{text}}

{{text}} {{text}} {{text}}

{{text}} {{text}} {{text}}

Address: {{address}}

Best regards,

{{name}}
{{address}}
{{phone_number}}
"""

pdf_file = FAKER.pdf_file(content=TEMPLATE, wrap_chars_after=80)



import os
import shutil
import tempfile
from typing import Callable, Optional, Union, overload

import xml2epub
from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import DEFAULT_FORMAT_FUNC, BytesValue, FileMixin, StringValue
from ..constants import DEFAULT_TEXT_MAX_NB_CHARS
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("EpubFileProvider",)


class EpubFileProvider(BaseProvider, FileMixin):
    """EPUB file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.epub_file import EpubFileProvider

        FAKER = Faker()
        FAKER.add_provider(EpubFileProvider)

        file = FAKER.epub_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.epub_file(
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.epub_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )
    """

    extension: str = "epub"

    @overload
    def epub_file(
        self: "EpubFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        title: Optional[str] = None,
        chapter_title: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def epub_file(
        self: "EpubFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        title: Optional[str] = None,
        chapter_title: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def epub_file(
        self: "EpubFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        title: Optional[str] = None,
        chapter_title: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a EPUB file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param max_nb_chars: Max number of chars for the content.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param title: E-book title. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param chapter_title: Chapter title. Might contain dynamic elements,
            which are then replaced by correspondent fixtures.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        content = self._generate_text_content(
            max_nb_chars=max_nb_chars,
            wrap_chars_after=wrap_chars_after,
            content=content,
            format_func=format_func,
        ).replace("\n", "<br>")

        title = self._generate_text_content(
            max_nb_chars=50,
            content=title,
            format_func=format_func,
        )

        chapter_title = self._generate_text_content(
            max_nb_chars=50,
            content=chapter_title,
            format_func=format_func,
        )

        _book = xml2epub.Epub(title)
        _chapter = xml2epub.create_chapter_from_string(
            content, title=chapter_title
        )
        _book.add_chapter(_chapter)
        _local_file_name = _book.create_epub(tempfile.gettempdir())

        data = {
            "content": content,
            "title": title,
            "chapter_title": chapter_title,
            "filename": filename,
            "storage": storage,
        }

        _raw_content = bytes()
        with open(_local_file_name, "rb") as fakefile:
            _raw_content = fakefile.read()
            os.remove(_local_file_name)
            shutil.rmtree(_book.EPUB_DIR)

        if raw:
            raw_content = BytesValue(_raw_content)
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, _raw_content)

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from faker import Faker
from faker_file.providers.helpers.inner import (
    create_inner_docx_file,
    create_inner_epub_file,
    create_inner_txt_file,
    fuzzy_choice_create_inner_file,
)
from faker_file.providers.zip_file import ZipFileProvider
from faker_file.storages.filesystem import FileSystemStorage

FAKER = Faker()
FAKER.add_provider(ZipFileProvider)

STORAGE = FileSystemStorage()

kwargs = {"storage": STORAGE, "generator": FAKER}

zip_file = FAKER.zip_file(
    prefix="zzz_archive_",
    options={
        "count": 50,
        "create_inner_file_func": fuzzy_choice_create_inner_file,
        "create_inner_file_args": {
            "func_choices": [
                (create_inner_docx_file, kwargs),
                (create_inner_epub_file, kwargs),
                (create_inner_txt_file, kwargs),
            ],
        },
        "directory": "zzz",
    },
)



# import mimetypes
import os
from email.message import EmailMessage
from email.policy import default
from typing import Any, Callable, Dict, Optional, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import (
    DEFAULT_FORMAT_FUNC,
    BytesValue,
    FileMixin,
    StringValue,
    returns_list,
)
from ..constants import DEFAULT_TEXT_MAX_NB_CHARS
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage
from .helpers.inner import create_inner_txt_file

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("EmlFileProvider",)


class EmlFileProvider(BaseProvider, FileMixin):
    """EML file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.eml_file import EmlFileProvider

        FAKER = Faker()
        FAKER.add_provider(EmlFileProvider)

        file = FAKER.eml_file()

    Usage example with attachments:

    .. code-block:: python

        from faker_file.providers.helpers.inner import create_inner_docx_file

        file = FAKER.eml_file(
            prefix="zzz_email_",
            options={
                "count": 5,
                "create_inner_file_func": create_inner_docx_file,
                "create_inner_file_args": {
                    "prefix": "zzz_docx_file_",
                    "max_nb_chars": 1_024,
                },
            }
        )

    Usage example of nested EMLs attachments:

    .. code-block:: python

        from faker_file.providers.helpers.inner import create_inner_eml_file

        file = FAKER.eml_file(
            options={
                "create_inner_file_func": create_inner_eml_file,
                "create_inner_file_args": {
                    "options": {
                        "create_inner_file_func": create_inner_docx_file,
                    }
                }
            }
        )

    If you want to see, which files were included inside the EML, check
    the ``file.data["files"]``.
    """

    extension: str = "eml"

    @overload
    def eml_file(
        self: "EmlFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        subject: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def eml_file(
        self: "EmlFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        subject: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def eml_file(
        self: "EmlFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        subject: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate an EML file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param options: Options (non-structured) for complex types, such as ZIP.
        :param max_nb_chars: Max number of chars for the content.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param subject: Email subject. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )
        fs_storage = FileSystemStorage()

        content = self._generate_text_content(
            max_nb_chars=max_nb_chars,
            wrap_chars_after=wrap_chars_after,
            content=content,
            format_func=format_func,
        )
        subject = self._generate_text_content(
            max_nb_chars=20,
            wrap_chars_after=wrap_chars_after,
            content=subject,
            format_func=format_func,
        )
        data: Dict[str, Any] = {
            "content": f"{subject}\n {content}",
            "inner": {},
            "filename": filename,
            "storage": storage,
        }

        msg = EmailMessage()
        msg["To"] = self.generator.email()
        msg["From"] = self.generator.email()
        msg["Subject"] = subject
        msg.set_content(content)
        data.update(
            {
                "to": msg["To"],
                "from": msg["From"],
                "subject": msg["Subject"],
                "body": content,
            }
        )

        # Specific
        if options:
            """
            A complex case. Could be initialized as follows:

            .. code-block:: python

                eml_file = FAKER.eml_file(
                    prefix="zzz_email_",
                    options={
                        "count": 5,
                        "create_inner_file_func": create_inner_docx_file,
                        "create_inner_file_args": {
                            "prefix": "zzz_file_",
                            "max_nb_chars": 1_024,
                            "content": "{{date}}\r\n{{text}}\r\n{{name}}",
                        },
                    }
                )
            """
            _count = options.get("count", 5)
            _create_inner_file_func = options.get(
                "create_inner_file_func", create_inner_txt_file
            )
            _create_inner_file_args = options.get("create_inner_file_args", {})

        else:
            # Defaults
            _count = 0
            _create_inner_file_func = create_inner_txt_file
            _create_inner_file_args = {}

        _kwargs = {"generator": self.generator}
        _kwargs.update(_create_inner_file_args)

        # If _create_inner_file_func returns a list of values
        if returns_list(_create_inner_file_func):
            _files = _create_inner_file_func(
                storage=fs_storage,
                **_kwargs,
            )
            for __file in _files:
                data["inner"][str(__file)] = __file
                __file_abs_path = fs_storage.abspath(__file)
                # _content_type, _encoding = mimetypes.guess_type(
                #     __file_abs_path
                # )
                # if _content_type is None or _encoding is not None:
                #     # No guess could be made, or the file is
                #     # encoded (compressed), so use a generic bag-of-bits
                #     # type.
                #     _content_type = "application/octet-stream"
                _content_type = "application/octet-stream"
                _maintype, _subtype = _content_type.split("/", 1)
                with open(__file_abs_path, "rb") as _fp:
                    _file_data = _fp.read()
                    msg.add_attachment(
                        _file_data,
                        maintype=_maintype,
                        subtype=_subtype,
                        filename=os.path.basename(__file),
                    )
                os.remove(__file_abs_path)  # Clean up temporary files
        # If _create_inner_file_func returns a single value
        else:
            for __i in range(_count):
                __file = _create_inner_file_func(
                    storage=fs_storage,
                    **_kwargs,
                )
                data["inner"][str(__file)] = __file
                __file_abs_path = fs_storage.abspath(__file)
                # _content_type, _encoding = mimetypes.guess_type(
                #     __file_abs_path
                # )
                # if _content_type is None or _encoding is not None:
                #     # No guess could be made, or the file is
                #     # encoded (compressed), so use a generic bag-of-bits
                #     # type.
                #     _content_type = "application/octet-stream"
                _content_type = "application/octet-stream"
                _maintype, _subtype = _content_type.split("/", 1)
                with open(__file_abs_path, "rb") as _fp:
                    _file_data = _fp.read()
                    msg.add_attachment(
                        _file_data,
                        maintype=_maintype,
                        subtype=_subtype,
                        filename=os.path.basename(__file),
                    )
                os.remove(__file_abs_path)  # Clean up temporary files

        if raw:
            raw_content = BytesValue(msg.as_bytes(policy=default))
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, msg.as_bytes(policy=default))

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    overload,
)

from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import DEFAULT_FORMAT_FUNC, BytesValue, StringValue
from ..constants import DEFAULT_IMAGE_MAX_NB_CHARS
from ..storages.base import BaseStorage
from .base.image_generator import BaseImageGenerator
from .mixins.graphic_image_mixin import GraphicImageMixin
from .mixins.image_mixin import DEFAULT_IMAGE_GENERATOR, ImageMixin

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "GraphicWebpFileProvider",
    "WebpFileProvider",
)


class WebpFileProvider(BaseProvider, ImageMixin):
    """WEBP file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.webp_file import WebpFileProvider

        FAKER = Faker()
        FAKER.add_provider(WebpFileProvider)

        file = FAKER.webp_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.webp_file(
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.webp_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )
    """

    extension: str = "webp"
    image_format: str = "webp"  # For Pillow

    @overload
    def webp_file(
        self: "WebpFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def webp_file(
        self: "WebpFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def webp_file(
        self: "WebpFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a WEBP file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param max_nb_chars: Max number of chars for the content.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param image_generator_cls: Image generator class.
        :param image_generator_kwargs: Image generator kwargs.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        return self._image_file(
            storage=storage,
            basename=basename,
            prefix=prefix,
            max_nb_chars=max_nb_chars,
            wrap_chars_after=wrap_chars_after,
            content=content,
            image_generator_cls=image_generator_cls,
            image_generator_kwargs=image_generator_kwargs,
            format_func=format_func,
            raw=raw,
            **kwargs,
        )


class GraphicWebpFileProvider(BaseProvider, GraphicImageMixin):
    """Graphic WEBP file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.webp_file import GraphicWebpFileProvider

        FAKER = Faker()
        FAKER.add_provider(GraphicWebpFileProvider)

        file = FAKER.graphic_webp_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.graphic_webp_file(
            prefix="zzz",
            size=(800, 800),
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.graphic_webp_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            basename="yyy",
            size=(1024, 1024),
        )
    """

    extension: str = "webp"
    image_format: str = "webp"

    @overload
    def graphic_webp_file(
        self: "GraphicWebpFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        size: Tuple[int, int] = (256, 256),
        hue: Union[int, Sequence[int], str, None] = None,
        luminosity: Optional[str] = None,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def graphic_webp_file(
        self: "GraphicWebpFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        size: Tuple[int, int] = (256, 256),
        hue: Union[int, Sequence[int], str, None] = None,
        luminosity: Optional[str] = None,
        **kwargs,
    ) -> StringValue: ...

    def graphic_webp_file(
        self: "GraphicWebpFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        size: Tuple[int, int] = (256, 256),
        hue: Union[int, Sequence[int], str, None] = None,
        luminosity: Optional[str] = None,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a graphic WEBP file with random lines.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param size: Image size in pixels.
        :param hue: Read more about
            ://faker.readthedocs.io/en/dev/providers/faker.providers.color.html
        :param luminosity: If given, the output string would be separated
             by line breaks after the given position.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        return self._image_file(
            image_format=self.image_format,
            storage=storage,
            basename=basename,
            prefix=prefix,
            size=size,
            hue=hue,
            luminosity=luminosity,
            raw=raw,
            **kwargs,
        )



from faker import Faker
from faker_file.providers.pdf_file import PdfFileProvider
from faker_file.providers.pdf_file.generators.pil_generator import (
    PilPdfGenerator,
)

FAKER = Faker()
FAKER.add_provider(PdfFileProvider)

pdf_file = FAKER.pdf_file(
    pdf_generator_cls=PilPdfGenerator,
    pdf_generator_kwargs={
        "encoding": "utf8",
        "font_size": 14,
        "page_width": 800,
        "page_height": 1200,
        "line_height": 16,
        "spacing": 5,
    },
    wrap_chars_after=100,
)



import argparse
import sys
import typing

from .. import __version__

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("main",)


def main():
    try:
        from .helpers import (
            PROVIDERS,
            generate_completion_file,
            generate_file,
            get_method_kwargs,
            is_optional_type,
        )
    except ImportError:
        print("You need to pip install faker-file[common] to use the CLI")
        sys.exit(1)

    parser = argparse.ArgumentParser(
        description="CLI for the faker-file package."
    )
    subparsers = parser.add_subparsers(
        dest="command", help="Available file providers."
    )

    # Add generate-completion subparser
    __generate_completion_subparser = subparsers.add_parser(
        "generate-completion",
        help="Generate bash completion file.",
    )

    # Add version subparser
    __version_subparser = subparsers.add_parser(
        "version",
        help="Print version.",
    )

    for method_name, provider in PROVIDERS.items():
        subparser = subparsers.add_parser(
            method_name,
            help=f"Generate a {method_name.split('_file')[0]} file.",
        )
        method_kwargs, annotations = get_method_kwargs(provider, method_name)
        for arg, default in method_kwargs.items():
            arg_type = annotations[arg]
            arg_kwargs = {
                "default": default,
                "help": f"{arg} (default: {default})",
                "type": (
                    arg_type.__args__[0]
                    if isinstance(arg_type, typing._GenericAlias)
                    and is_optional_type(arg_type)
                    else arg_type
                ),
            }

            subparser.add_argument(f"--{arg}", **arg_kwargs)

        # Add the optional num_files argument
        subparser.add_argument(
            "--nb_files",
            default=1,
            type=int,
            help="number of files to generate (default: 1)",
        )

    args = parser.parse_args()

    if args.command == "generate-completion":
        generate_completion_file()
    elif args.command == "version":
        print(__version__)
    elif args.command:
        kwargs = {k: v for k, v in vars(args).items() if k not in ("command",)}
        for counter in range(args.nb_files):
            output_file = generate_file(args.command, **kwargs)
            print(
                f"Generated {args.command} file "
                f"({counter+1} of {args.nb_files}): "
                f"{output_file.data['filename']}"
            )
    else:
        parser.print_help()
        sys.exit(1)



# Required imports
from faker import Faker
from faker_file.base import DynamicTemplate
from faker_file.contrib.pdf_file.reportlab_snippets import (
    add_page_break,
    add_paragraph,
    add_picture,
    add_table,
)
from faker_file.providers.pdf_file import PdfFileProvider
from faker_file.providers.pdf_file.generators.reportlab_generator import (
    ReportlabPdfGenerator,
)

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(PdfFileProvider)  # Register provider

# Create a PDF file with paragraph, picture, table and manual page breaks
# in between the mentioned elements. The ``DynamicTemplate`` simply
# accepts a list of callables (such as ``add_paragraph``,
# ``add_page_break``) and dictionary to be later on fed to the callables
# as keyword arguments for customising the default values.
pdf_file = FAKER.pdf_file(
    pdf_generator_cls=ReportlabPdfGenerator,
    content=DynamicTemplate(
        [
            (add_paragraph, {}),  # Add paragraph
            (add_page_break, {}),  # Add page break
            (add_picture, {}),  # Add picture
            (add_page_break, {}),  # Add page break
            (add_table, {}),  # Add table
            (add_page_break, {}),  # Add page break
        ]
    ),
)

# You could make the list as long as you like or simply multiply for
# easier repetition as follows:
pdf_file = FAKER.pdf_file(
    pdf_generator_cls=ReportlabPdfGenerator,
    content=DynamicTemplate(
        [
            (add_paragraph, {}),  # Add paragraph
            (add_page_break, {}),  # Add page break
            (add_picture, {}),  # Add picture
            (add_page_break, {}),  # Add page break
            (add_table, {}),  # Add table
            (add_page_break, {}),  # Add page break
        ]
        * 5  # Will repeat your config 5 times
    ),
)



from io import BytesIO

from PIL import Image as PilImage
from reportlab.lib import colors
from reportlab.lib.styles import ParagraphStyle, getSampleStyleSheet
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.platypus import Image, PageBreak, Paragraph, Table, TableStyle

from ...base import DEFAULT_FORMAT_FUNC

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "add_h1_heading",
    "add_h2_heading",
    "add_h3_heading",
    "add_h4_heading",
    "add_h5_heading",
    "add_h6_heading",
    "add_heading",
    "add_page_break",
    "add_paragraph",
    "add_picture",
    "add_table",
)


def add_table(
    provider,
    generator,
    story,
    data,
    counter,
    **kwargs,
):
    """Add table function."""
    rows = kwargs.get("rows", 3)
    cols = kwargs.get("cols", 4)
    font_name = kwargs.get("font_name", "Helvetica-Bold")
    font_size = kwargs.get("font_size", 14)
    align = kwargs.get("align", "CENTER")

    # Define your table headers
    headers = [f"Header {i + 1}" for i in range(cols)]

    # Generate the rest of the table data
    table_data = [
        [provider.generator.word() for _ in range(cols)] for _ in range(rows)
    ]

    # Add the headers to the table data
    table_data.insert(0, headers)

    # Create the table object
    table = Table(table_data)

    # Apply table styles
    table.setStyle(
        TableStyle(
            [
                ("BACKGROUND", (0, 0), (-1, 0), colors.grey),
                ("TEXTCOLOR", (0, 0), (-1, 0), colors.whitesmoke),
                ("ALIGN", (0, 0), (-1, -1), align),
                ("FONTNAME", (0, 0), (-1, 0), font_name),
                ("FONTSIZE", (0, 0), (-1, 0), font_size),
                ("BOTTOMPADDING", (0, 0), (-1, 0), 12),
                ("BACKGROUND", (0, 1), (-1, -1), colors.beige),
                ("GRID", (0, 0), (-1, -1), 1, colors.black),
                ("BOX", (0, 0), (-1, -1), 1, colors.black),
            ]
        )
    )

    # Add the table to the document and build it
    story.append(table)


def add_picture(
    provider,
    generator,
    story,
    data,
    counter,
    **kwargs,
):
    """Add picture function."""
    image = kwargs.get("image", provider.generator.image())

    # Create a BytesIO object and load the image data
    with BytesIO(image) as input_stream:
        pil_image = PilImage.open(input_stream)

        # Resize the image
        new_width = 400
        new_height = 400
        pil_image = pil_image.resize((new_width, new_height))

        # Create a BytesIO object outside the 'with' statement
        output_stream = BytesIO()
        pil_image.save(output_stream, format="JPEG")
        output_stream.seek(0)  # Move to the start of the stream

        # Now you can use output_stream as your image data
        img = Image(output_stream)
        img.width = new_width
        img.height = new_height
        story.append(img)


def add_page_break(
    provider,
    generator,
    story,
    data,
    counter,
    **kwargs,
):
    """Add page break function."""
    # Insert a page break
    story.append(PageBreak())


def add_paragraph(
    provider,
    generator,
    story,
    data,
    counter,
    **kwargs,
):
    """Add paragraph function."""
    content = kwargs.get("content", None)
    max_nb_chars = kwargs.get("content", 5_000)
    wrap_chars_after = kwargs.get("wrap_chars_after", None)
    format_func = kwargs.get("format_func", DEFAULT_FORMAT_FUNC)

    _content = provider._generate_text_content(
        max_nb_chars=max_nb_chars,
        wrap_chars_after=wrap_chars_after,
        content=content,
        format_func=format_func,
    )

    # Insert a paragraph
    styles = getSampleStyleSheet()
    style_paragraph = styles["Normal"]
    style_paragraph.fontName = generator.font_name
    pdfmetrics.registerFont(TTFont(generator.font_name, generator.font_path))
    content = provider.generator.text(max_nb_chars=5_000)
    paragraph = Paragraph(content, style_paragraph)
    story.append(paragraph)

    # Meta-data
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_paragraph", {})
    data["content_modifiers"]["add_paragraph"].setdefault(counter, [])
    data["content_modifiers"]["add_paragraph"][counter].append(_content)
    data["content"] += "\r\n" + _content


def add_heading(
    provider,
    generator,
    story,
    data,
    counter,
    **kwargs,
):
    """Add heading function."""
    content = kwargs.get("content", None)
    max_nb_chars = kwargs.get("max_nb_chars", 30)
    wrap_chars_after = kwargs.get("wrap_chars_after", None)
    format_func = kwargs.get("format_func", DEFAULT_FORMAT_FUNC)
    level = kwargs.get("level", 1)
    if level < 1 or level > 6:
        level = 1

    _content = provider._generate_text_content(
        max_nb_chars=max_nb_chars,
        wrap_chars_after=wrap_chars_after,
        content=content,
        format_func=format_func,
    )

    # Insert a heading
    styles = getSampleStyleSheet()

    # Define the heading style based on the level
    heading_style = ParagraphStyle(
        f"Heading{level}",
        parent=styles[f"Heading{level}"],
        fontSize=14 - level,
        spaceAfter=12,
    )

    heading_style.fontName = generator.font_name
    pdfmetrics.registerFont(TTFont(generator.font_name, generator.font_path))

    heading = Paragraph(_content, heading_style)
    story.append(heading)

    # Meta-data
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_heading", {})
    data["content_modifiers"]["add_heading"].setdefault(counter, [])
    data["content_modifiers"]["add_heading"][counter].append(_content)
    data["content"] += "\r\n" + _content


def add_h1_heading(provider, generator, document, data, counter, **kwargs):
    """Callable responsible for the h1 heading generation."""
    return add_heading(
        provider, generator, document, data, counter, level=1, **kwargs
    )


def add_h2_heading(provider, generator, document, data, counter, **kwargs):
    """Callable responsible for the h2 heading generation."""
    return add_heading(
        provider, generator, document, data, counter, level=2, **kwargs
    )


def add_h3_heading(provider, generator, document, data, counter, **kwargs):
    """Callable responsible for the h3 heading generation."""
    return add_heading(
        provider, generator, document, data, counter, level=3, **kwargs
    )


def add_h4_heading(provider, generator, document, data, counter, **kwargs):
    """Callable responsible for the h4 heading generation."""
    return add_heading(
        provider, generator, document, data, counter, level=4, **kwargs
    )


def add_h5_heading(provider, generator, document, data, counter, **kwargs):
    """Callable responsible for the h5 heading generation."""
    return add_heading(
        provider, generator, document, data, counter, level=5, **kwargs
    )


def add_h6_heading(provider, generator, document, data, counter, **kwargs):
    """Callable responsible for the h6 heading generation."""
    return add_heading(
        provider, generator, document, data, counter, level=6, **kwargs
    )



import random
import string
from typing import Any, Optional

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("BaseStorage",)


class BaseStorage:
    """Base storage."""

    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs

    def generate_basename(
        self: "BaseStorage",
        prefix: str = "tmp",
        length: int = 8,
    ) -> str:
        """Generate a random alphanumeric sequence."""
        if not prefix:
            prefix = "tmp"
        # Use lowercase letters, digits and underscore
        characters = string.ascii_lowercase + string.digits + "_"
        return prefix + "".join(random.choices(characters, k=length))

    def generate_filename(
        self: "BaseStorage",
        extension: str,
        prefix: Optional[str] = None,
        basename: Optional[str] = None,
    ) -> Any:
        """Generate filename."""
        raise NotImplementedError(
            "Method generate_filename is not implemented!"
        )

    def write_text(
        self: "BaseStorage",
        filename: Any,
        data: str,
        encoding: Optional[str] = None,
    ) -> int:
        """Write text."""
        raise NotImplementedError("Method write_text is not implemented!")

    def write_bytes(self: "BaseStorage", filename: Any, data: bytes) -> int:
        """Write bytes."""
        raise NotImplementedError("Method write_bytes is not implemented!")

    def exists(self: "BaseStorage", filename: Any) -> bool:
        """Check if file exists."""
        raise NotImplementedError("Method exists is not implemented!")

    def relpath(self: "BaseStorage", filename: Any) -> str:
        """Return relative path."""
        raise NotImplementedError("Method relpath is not implemented!")

    def abspath(self: "BaseStorage", filename: Any) -> str:
        """Return absolute path."""
        raise NotImplementedError("Method abspath is not implemented!")

    def unlink(self: "BaseStorage", filename: Any) -> None:
        """Delete the file."""
        raise NotImplementedError("Method unlink is not implemented!")



from pathlib import Path
from typing import Optional, Union, overload

from faker.providers import BaseProvider

from ..base import BytesValue, FileMixin, StringValue
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("FileFromPathProvider",)


class FileFromPathProvider(BaseProvider, FileMixin):
    """File from given path provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.file_from_path import (
            FileFromPathProvide
        )

        FAKER = Faker()
        FAKER.add_provider(FileFromPathProvider)

        file = FAKER.file_from_path(
            path="/path/to/file.pdf"
        )

    Usage example with options:

    .. code-block:: python

        file = FAKER.file_from_path(
            path="/path/to/file.pdf",
            prefix="zzz",
        )
    """

    extension: str = ""

    @overload
    def file_from_path(
        self: "FileFromPathProvider",
        path: str,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def file_from_path(
        self: "FileFromPathProvider",
        path: str,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        **kwargs,
    ) -> StringValue: ...

    def file_from_path(
        self: "FileFromPathProvider",
        path: str,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """File from given path.

        :param path: Path to source file.
        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        # Specific
        source_file = Path(path)

        # Generic
        filename = storage.generate_filename(
            extension=source_file.suffix[1:],
            prefix=prefix,
            basename=basename,
        )
        data = {"filename": filename, "storage": storage}

        # Specific
        with open(path, "rb") as _file:
            if raw:
                raw_content = BytesValue(_file.read())
                raw_content.data = data
                return raw_content

            storage.write_bytes(filename, _file.read())

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from faker import Faker
from faker_file.providers.txt_file import TxtFileProvider
from faker_file.storages.aws_s3 import AWSS3Storage

FAKER = Faker()
FAKER.add_provider(TxtFileProvider)

AWS_S3_STORAGE = AWSS3Storage(
    bucket_name="your-bucket-name",
    root_path="",
    rel_path="user/uploads",
)

txt_file = FAKER.txt_file(storage=AWS_S3_STORAGE)



from faker import Faker
from faker_file.providers.image.weasyprint_generator import (
    WeasyPrintImageGenerator,
)
from faker_file.providers.png_file import PngFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(PngFileProvider)  # Register provider

# Generate image file using `WeasyPrint`
png_file = FAKER.png_file(image_generator_cls=WeasyPrintImageGenerator)



import logging
import textwrap
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, List, Type, Union

from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider
from PIL import Image, ImageDraw, ImageFont

from ....base import DynamicTemplate
from ....constants import DEFAULT_FILE_ENCODING
from ...base.pdf_generator import BasePdfGenerator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("PilPdfGenerator",)

LOGGER = logging.getLogger(__name__)


class PilPdfGenerator(BasePdfGenerator):
    """Pillow PDF generator.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.pdf_file import PdfFileProvider
        from faker_file.providers.pdf_file.generators import pil_generator

        FAKER = Faker()
        FAKER.add_provider(PdfFileProvider)

        file = FAKER.pdf_file(
            pdf_generator_cls=pil_generator.PilPdfGenerator
        )

    With options:

    .. code-bloock:: python

        file = FAKER.pdf_file(
            pdf_generator_cls=PilPdfGenerator,
            pdf_generator_kwargs={
                "encoding": "utf8",
                "font_size": 14,
                "page_width": 800,
                "page_height": 1200,
                "line_height": 16,
                "spacing": 5,
            },
            wrap_chars_after=100,
        )
    """

    encoding: str = DEFAULT_FILE_ENCODING
    font: str = str(Path("Pillow") / "Tests" / "fonts" / "DejaVuSans.ttf")
    font_size: int = 12
    page_width: int = 794  # A4 size at 96 DPI
    page_height: int = 1123  # A4 size at 96 DPI
    line_height: int = 14
    spacing: int = 6  # Letter spacing

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pages = []
        self.img = None
        self.draw = None
        self.image_mode = "RGB"

    @classmethod
    def find_max_fit_for_multi_line_text(
        cls: Type["PilPdfGenerator"],
        draw: ImageDraw,
        lines: List[str],
        font: ImageFont,
        max_width: int,
    ):
        # Find the longest line
        longest_line = str(max(lines, key=len))
        return cls.find_max_fit_for_single_line_text(
            draw, longest_line, font, max_width
        )

    @classmethod
    def find_max_fit_for_single_line_text(
        cls: Type["PilPdfGenerator"],
        draw: "ImageDraw",
        text: str,
        font: ImageFont,
        max_width: int,
    ) -> int:
        low, high = 0, len(text)
        while low < high:
            mid = (high + low) // 2
            text_width, _ = draw.textsize(text[:mid], font=font)

            if text_width > max_width:
                high = mid
            else:
                low = mid + 1

        return low - 1

    def handle_kwargs(self: "PilPdfGenerator", **kwargs) -> None:
        """Handle kwargs."""
        if "encoding" in kwargs:
            self.encoding = kwargs["encoding"]
        if "font_size" in kwargs:
            self.font_size = kwargs["font_size"]
        if "page_width" in kwargs:
            self.page_width = kwargs["page_width"]
        if "page_height" in kwargs:
            self.page_height = kwargs["page_height"]
        if "line_height" in kwargs:
            self.line_height = kwargs["line_height"]
        if "spacing" in kwargs:
            self.spacing = kwargs["spacing"]
        if "image_mode" in kwargs:
            self.image_mode = kwargs["image_mode"]

    def create_image_instance(self) -> Image:
        return Image.new(
            self.image_mode,
            (self.page_width, self.page_height),
            (255, 255, 255),
        )

    def start_new_page(self):
        self.img = self.create_image_instance()
        self.draw = ImageDraw.Draw(self.img)

    def save_and_start_new_page(self):
        self.pages.append(self.img.copy())
        self.start_new_page()

    def generate(
        self: "PilPdfGenerator",
        content: Union[str, DynamicTemplate],
        data: Dict[str, Any],
        provider: Union[Faker, Generator, Provider],
        **kwargs,
    ) -> bytes:
        """Generate PDF."""
        position = (0, 0)
        if isinstance(content, DynamicTemplate):
            self.start_new_page()
            for counter, (ct_modifier, ct_modifier_kwargs) in enumerate(
                content.content_modifiers
            ):
                ct_modifier_kwargs["position"] = position
                # LOGGER.debug(f"ct_modifier_kwargs: {ct_modifier_kwargs}")
                add_page, position = ct_modifier(
                    provider,
                    self,
                    data,
                    counter,
                    **ct_modifier_kwargs,
                )

            self.pages.append(self.img.copy())  # Add as a new page
        else:
            self.img = self.create_image_instance()
            self.draw = ImageDraw.Draw(self.img)
            font = ImageFont.truetype(self.font, self.font_size)

            # The `content_specs` is a dictionary that holds two keys:
            # `max_nb_chars` and `wrap_chars_after`. Those are the same values
            # passed to the `PdfFileProvider`.
            content_specs = kwargs.get("content_specs", {})
            lines = content.split("\n")
            line_max_num_chars = self.find_max_fit_for_multi_line_text(
                self.draw,
                lines,
                font,
                self.page_width,
            )
            wrap_chars_after = content_specs.get("wrap_chars_after")
            if (
                not wrap_chars_after
                or wrap_chars_after
                and (wrap_chars_after > line_max_num_chars)
            ):
                lines = textwrap.wrap(content, line_max_num_chars)

            y_text = 0
            for counter, line in enumerate(lines):
                text_width, text_height = self.draw.textsize(
                    line, font=font, spacing=self.spacing
                )
                # if counter % max_lines_per_page == 0:
                if y_text + text_height > self.page_height:
                    self.save_and_start_new_page()
                    y_text = 0

                self.draw.text(
                    (0, y_text),
                    line,
                    fill=(0, 0, 0),
                    spacing=self.spacing,
                    font=font,
                )
                y_text += text_height + self.line_height

            self.pages.append(self.img.copy())  # Add as a new page

        buffer = BytesIO()
        # Save as multi-page PDF
        self.pages[0].save(
            buffer, save_all=True, append_images=self.pages[1:], format="PDF"
        )

        return buffer.getvalue()



import importlib
import random
from textwrap import wrap
from typing import Any, Type

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "load_class_from_path",
    "random_pop",
    "wrap_text",
)


def wrap_text(text: str, wrap_chars_after: int) -> str:
    return "\n".join(
        wrap(
            text=text,
            width=wrap_chars_after,
            replace_whitespace=False,
            # drop_whitespace=False,
        )
    )


def load_class_from_path(full_path: str) -> Type:
    """Load a class from a given full path string identifier.

    :param full_path: The full path to the class,
        e.g. 'module.submodule.MyClass'.
    :return: The loaded class.
    :raise: If the module cannot be found or the class does
        not exist in the module, it raises ImportError.

    Usage example:

    .. code-block:: python

        my_class = load_class_from_path("module.submodule.MyClass")
        instance = my_class()
    """
    try:
        module_name, class_name = full_path.rsplit(".", 1)
        module = importlib.import_module(module_name)

        if not hasattr(module, class_name):
            raise ImportError(
                f"Class '{class_name}' not found in module '{module_name}'"
            )

        loaded_class = getattr(module, class_name)

        if not isinstance(loaded_class, type):
            raise ImportError(f"'{full_path}' does not point to a class")

        return loaded_class
    except ImportError as e:
        raise ImportError(f"Error loading class from path '{full_path}': {e}")


def random_pop(lst: list) -> Any:
    """Randomly pops element from the given list. Alters the list.

    :param lst: List to pop element from.
    :return: A single element from the list.

    Usage example:

    .. code-block:: python

        from faker_file.helpers import random_pop
        my_list = [1, 2, 3, 4, 5]
        element = random_pop(my_list)
    """
    if len(lst) > 0:
        idx = random.randrange(len(lst))
        return lst.pop(idx)
    else:
        return None



from faker import Faker
from faker_file.providers.txt_file import TxtFileProvider
from faker_file.storages.sftp_storage import SFTPStorage

FAKER = Faker()
FAKER.add_provider(TxtFileProvider)

SFTP_STORAGE = SFTPStorage(
    host="your-sftp-host.domain",
    port=22,
    username="your-sftp-username",
    password="your-sftp-password",
    root_path="/dir-name",
)

# txt_file = FAKER.txt_file(storage=SFTP_STORAGE)



from django.conf import settings
from factory import Faker
from factory.django import DjangoModelFactory
from faker_file.providers.pdf_file import PdfFileProvider
from faker_file.storages.aws_s3 import AWSS3Storage

from upload.models import Upload

STORAGE = AWSS3Storage(
    bucket_name=settings.AWS_STORAGE_BUCKET_NAME,
    root_path="",
    rel_path="",
)

Faker.add_provider(PdfFileProvider)


class UploadFactory(DjangoModelFactory):
    name = Faker("word")
    description = Faker("text")
    file = Faker("pdf_file", storage=STORAGE)

    class Meta:
        model = Upload


# Usage example
upload = UploadFactory()



from typing import Callable, Dict, Optional, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider
from tablib import Dataset

from ...base import DEFAULT_FORMAT_FUNC, BytesValue, FileMixin, StringValue
from ...registry import FILE_REGISTRY
from ...storages.base import BaseStorage
from ...storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("TabularDataMixin",)


class TabularDataMixin(FileMixin):
    """Tabular data mixin."""

    @overload
    def _tabular_data_file(
        self: "TabularDataMixin",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        data_columns: Optional[Dict[str, str]] = None,
        num_rows: int = 10,
        content: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def _tabular_data_file(
        self: "TabularDataMixin",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        data_columns: Optional[Dict[str, str]] = None,
        num_rows: int = 10,
        content: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def _tabular_data_file(
        self: "TabularDataMixin",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        data_columns: Optional[Dict[str, str]] = None,
        num_rows: int = 10,
        content: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a tabular data file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param data_columns: The ``data_columns`` argument expects a list or a
            tuple of string tokens, and these string tokens will be passed to
            :meth:`pystr_format()
            <faker.providers.python.Provider.pystr_format>`
            for data generation. Argument Groups are used to pass arguments
            to the provider methods. Both ``header`` and ``data_columns`` must
            be of the same length.
        :param num_rows: The ``num_rows`` argument controls how many rows of
            data to generate, and the ``include_row_ids`` argument may be set
            to ``True`` to include a sequential row ID column.
        :param content: List of dicts with content (JSON-like format).
            If given, used as is.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        if self.generator is None:
            self.generator = Faker()

        # Specific
        if content is None:
            default_data_columns = {
                "name": "{{name}}",
                "residency": "{{address}}",
            }
            data_columns = (
                data_columns if data_columns else default_data_columns
            )
            content = self.generator.json(
                data_columns=data_columns,
                num_rows=num_rows,
            )

        data = {"content": content, "filename": filename, "storage": storage}

        dataset = Dataset()
        dataset.load(content, format="json")

        _raw_content = dataset.export(self.extension)

        if raw:
            raw_content = BytesValue(_raw_content)
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, _raw_content)

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from io import BytesIO

from docx.enum.text import WD_PARAGRAPH_ALIGNMENT

from ..base import DEFAULT_FORMAT_FUNC

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "add_h1_heading",
    "add_h2_heading",
    "add_h3_heading",
    "add_h4_heading",
    "add_h5_heading",
    "add_h6_heading",
    "add_heading",
    "add_page_break",
    "add_paragraph",
    "add_picture",
    "add_table",
    "add_title_heading",
)


def add_table(provider, document, data, counter, **kwargs):
    """Callable responsible for the table generation."""
    table = document.add_table(
        kwargs.get("rows", 3),
        kwargs.get("cols", 4),
    )

    # Modifications of `data` is not required for generation
    # of the file, but is useful for when you want to get
    # the text content of the file.
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_table", {})
    data["content_modifiers"]["add_table"].setdefault(counter, [])

    for row in table.rows:
        for cell in row.cells:
            text = provider.generator.paragraph()
            cell.text = text
            # Useful when you want to get the text content of the file.
            data["content_modifiers"]["add_table"][counter].append(text)
            data["content"] += "\r\n" + text


def add_picture(provider, document, data, counter, **kwargs):
    """Callable responsible for the picture generation."""
    image = kwargs.get("image", provider.generator.image())
    document.add_picture(BytesIO(image))

    # # Modifications of `data` is not required for generation
    # # of the file, but is useful for when you want to get
    # # the text content of the file.
    # data.setdefault("content_modifiers", {})
    # data["content_modifiers"].setdefault("add_picture", {})
    # data["content_modifiers"]["add_picture"].setdefault(counter, [])
    # data["content_modifiers"]["add_picture"][counter].append(
    #     jpeg_file.data["content"]
    # )
    # data["content"] += "\r\n" + jpeg_file.data["content"]


def add_page_break(provider, document, data, counter, **kwargs):
    """Callable responsible for page break generation."""
    # Insert a page break
    document.add_page_break()


def add_paragraph(provider, document, data, counter, **kwargs):
    """Callable responsible for the paragraph generation."""
    content = kwargs.get("content", None)
    max_nb_chars = kwargs.get("content", 5_000)
    wrap_chars_after = kwargs.get("wrap_chars_after", None)
    format_func = kwargs.get("format_func", DEFAULT_FORMAT_FUNC)

    _content = provider._generate_text_content(
        max_nb_chars=max_nb_chars,
        wrap_chars_after=wrap_chars_after,
        content=content,
        format_func=format_func,
    )
    document.add_paragraph(_content)

    # Meta-data
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_paragraph", {})
    data["content_modifiers"]["add_paragraph"].setdefault(counter, [])
    data["content_modifiers"]["add_paragraph"][counter].append(_content)
    data["content"] += "\r\n" + _content


def add_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the heading generation."""
    content = kwargs.get("content", None)
    max_nb_chars = kwargs.get("content", 30)
    wrap_chars_after = kwargs.get("wrap_chars_after", None)
    format_func = kwargs.get("format_func", DEFAULT_FORMAT_FUNC)
    level = kwargs.get("level", 0)

    _content = provider._generate_text_content(
        max_nb_chars=max_nb_chars,
        wrap_chars_after=wrap_chars_after,
        content=content,
        format_func=format_func,
    )

    heading = document.add_heading(level=level)
    heading.add_run(_content)

    # Aligning the heading to center
    heading.alignment = WD_PARAGRAPH_ALIGNMENT.CENTER

    # Meta-data
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_heading", {})
    data["content_modifiers"]["add_heading"].setdefault(counter, [])
    data["content_modifiers"]["add_heading"][counter].append(_content)
    data["content"] += "\r\n" + _content


def add_title_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the title heading generation."""
    return add_heading(provider, document, data, counter, level=0, **kwargs)


def add_h1_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the h1 heading generation."""
    return add_heading(provider, document, data, counter, level=1, **kwargs)


def add_h2_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the h2 heading generation."""
    return add_heading(provider, document, data, counter, level=2, **kwargs)


def add_h3_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the h3 heading generation."""
    return add_heading(provider, document, data, counter, level=3, **kwargs)


def add_h4_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the h4 heading generation."""
    return add_heading(provider, document, data, counter, level=4, **kwargs)


def add_h5_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the h5 heading generation."""
    return add_heading(provider, document, data, counter, level=5, **kwargs)


def add_h6_heading(provider, document, data, counter, **kwargs):
    """Callable responsible for the h6 heading generation."""
    return add_heading(provider, document, data, counter, level=6, **kwargs)



import json
import tempfile
import unittest
from pathlib import Path
from typing import Any, Dict, Type, Union

from faker import Faker
from parametrize import parametrize
from pathy import use_fs, use_fs_cache

from ..providers.json_file import JsonFileProvider
from ..registry import FILE_REGISTRY
from ..storages.aws_s3 import AWSS3Storage
from ..storages.azure_cloud_storage import AzureCloudStorage
from ..storages.base import BaseStorage
from ..storages.cloud import CloudStorage, PathyFileSystemStorage
from ..storages.filesystem import FileSystemStorage
from ..storages.google_cloud_storage import GoogleCloudStorage
from .data import GCS_CREDENTIALS_JSON

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("TestStoragesTestCase",)

FAKER = Faker()
FAKER.add_provider(JsonFileProvider)
FS_STORAGE = FileSystemStorage()
GCS_CREDENTIALS_JSON_FILENAME = FS_STORAGE.generate_filename(
    basename="gc_credentials", extension="json"
)


class TestStoragesTestCase(unittest.TestCase):
    """Test storages."""

    def setUp(self) -> None:
        super().setUp()

        gcs_credentials_json_filename = Path(GCS_CREDENTIALS_JSON_FILENAME)
        FAKER.json_file(
            basename=gcs_credentials_json_filename.stem,
            content=json.dumps(GCS_CREDENTIALS_JSON),
        )

    def tearDown(self) -> None:
        super().tearDown()
        FILE_REGISTRY.clean_up()  # Clean up files

    @parametrize(
        "storage_cls, kwargs, prefix, basename, extension",
        [
            # FileSystemStorage
            (
                FileSystemStorage,
                {},
                "zzz",
                None,
                "docx",
            ),
            (
                FileSystemStorage,
                {},
                None,
                "my_zzz_filename",
                "docx",
            ),
            # PathyFileSystemStorage
            (
                PathyFileSystemStorage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                },
                "zzz",
                None,
                "docx",
            ),
            (
                PathyFileSystemStorage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                },
                None,
                "my_zzz_filename",
                "docx",
            ),
            # AWS S3
            (
                AWSS3Storage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                    "credentials": {
                        "key_id": "key",
                        "key_secret": "key_secret",
                    },
                },
                "zzz",
                None,
                "docx",
            ),
            (
                AWSS3Storage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                    "credentials": {
                        "key_id": "key",
                        "key_secret": "key_secret",
                    },
                },
                None,
                "my_zzz_filename",
                "docx",
            ),
            # Google Cloud Storage
            (
                GoogleCloudStorage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                    "credentials": {
                        "json_file_path": GCS_CREDENTIALS_JSON_FILENAME,
                    },
                },
                "zzz",
                None,
                "docx",
            ),
            (
                GoogleCloudStorage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                    "credentials": {
                        "json_file_path": GCS_CREDENTIALS_JSON_FILENAME,
                    },
                },
                None,
                "my_zzz_filename",
                "docx",
            ),
            # Azure Cloud Storage
            (
                AzureCloudStorage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                    "credentials": {"connection_string": "abcd1234"},
                },
                "zzz",
                None,
                "docx",
            ),
            (
                AzureCloudStorage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                    "credentials": {"connection_string": "abcd1234"},
                },
                None,
                "my_zzz_filename",
                "docx",
            ),
        ],
    )
    def test_storage(
        self: "TestStoragesTestCase",
        storage_cls: Type[BaseStorage],
        kwargs: Dict[str, Any],
        prefix: Union[str, None],
        basename: Union[str, None],
        extension: str,
    ) -> None:
        """Test storage."""
        # Just for testing purposes
        if issubclass(storage_cls, CloudStorage):
            use_fs(Path(tempfile.gettempdir()))
            use_fs_cache()

        storage = storage_cls(**kwargs)
        # Text file
        filename_text = storage.generate_filename(
            basename=basename, prefix=prefix, extension=extension
        )
        # Write to the text file
        text_result = storage.write_text(filename_text, "Lorem ipsum")
        # Check if file exists
        self.assertTrue(storage.exists(filename_text))
        # Assert correct return value
        self.assertIsInstance(text_result, int)
        # Clean up
        storage.unlink(filename_text)

        # Bytes
        filename_bytes = storage.generate_filename(
            basename=basename, prefix=prefix, extension=extension
        )
        # Write to bytes file
        bytes_result = storage.write_bytes(filename_bytes, b"Lorem ipsum")
        # Check if file exists
        self.assertTrue(storage.exists(filename_bytes))
        # Assert correct return value
        self.assertIsInstance(bytes_result, int)

        # Clean up
        storage.unlink(filename_bytes)

    @parametrize(
        "storage_cls, kwargs, prefix, extension",
        [
            # FileSystemStorage
            (
                FileSystemStorage,
                {},
                "zzz",
                "",
            ),
            # PathyFileSystemStorage
            (
                PathyFileSystemStorage,
                {
                    "bucket_name": "testing",
                    "rel_path": "tmp",
                },
                "zzz",
                "",
            ),
        ],
    )
    def test_storage_generate_filename_exceptions(
        self: "TestStoragesTestCase",
        storage_cls: Type[BaseStorage],
        kwargs: Dict[str, Any],
        prefix: str,
        extension: str,
    ) -> None:
        """Test storage `generate_filename` exceptions."""
        storage = storage_cls(**kwargs)

        with self.assertRaises(Exception):
            # Generate filename
            storage.generate_filename(prefix=prefix, extension=extension)

        with self.assertRaises(Exception):
            # Generate filename
            storage.generate_filename(basename=prefix, extension=extension)

    @parametrize(
        "storage_cls, kwargs",
        [
            # CloudStorage
            (CloudStorage, {"bucket_name": "testing"}),
        ],
    )
    def test_storage_initialization_exceptions(
        self: "TestStoragesTestCase",
        storage_cls: Type[BaseStorage],
        kwargs: Dict[str, Any],
    ) -> None:
        """Test storage initialization exceptions."""
        with self.assertRaises(Exception):
            # Initialize the storage
            storage_cls(**kwargs)

    @parametrize(
        "method_name, method_kwargs",
        [
            ("generate_filename", {"prefix": "zzz", "extension": "txt"}),
            (
                "generate_filename",
                {"basename": "my_zzz_file", "extension": "txt"},
            ),
            ("write_text", {"filename": "test.txt", "data": "Test"}),
            ("write_bytes", {"filename": "test.txt", "data": b"Test"}),
            ("exists", {"filename": "test.txt"}),
            ("relpath", {"filename": "test.txt"}),
            ("abspath", {"filename": "test.txt"}),
            ("unlink", {"filename": "test.txt"}),
        ],
    )
    def test_base_storage_exceptions(
        self: "TestStoragesTestCase",
        method_name: str,
        method_kwargs: Dict[str, Any],
    ) -> None:
        """Test Base storage exceptions."""
        base_storage = BaseStorage()
        method = getattr(base_storage, method_name)
        with self.assertRaises(NotImplementedError):
            method(**method_kwargs)

    @parametrize(
        "method_name, method_kwargs",
        [
            ("authenticate", {}),
        ],
    )
    def test_cloud_storage_exceptions(
        self: "TestStoragesTestCase",
        method_name: str,
        method_kwargs: Dict[str, Any],
    ) -> None:
        """Test Base storage exceptions."""

        class TestCloudStorage(CloudStorage):
            schema: str = "file"

        test_storage = TestCloudStorage(bucket_name="testing")  # type: ignore
        method = getattr(test_storage, method_name)
        with self.assertRaises(NotImplementedError):
            method(**method_kwargs)

    def test_file_system_storage_abspath(self: "TestStoragesTestCase") -> None:
        """Test `FileSystemStorage` `abspath`."""
        storage = FileSystemStorage(
            root_path=tempfile.gettempdir(),
            rel_path="rel_tmp",
        )
        filename = storage.generate_filename(prefix="", extension="tmp")
        self.assertTrue(storage.abspath(filename).startswith("/tmp/rel_tmp/"))

    def test_pathy_file_system_storage_abspath(
        self: "TestStoragesTestCase",
    ) -> None:
        """Test `PathyFileSystemStorage` `abspath`."""
        storage = PathyFileSystemStorage(
            bucket_name="faker-file-tmp",
            root_path="root_tmp",
            rel_path="rel_tmp",
        )
        filename = storage.generate_filename(prefix="", extension="tmp")
        self.assertTrue(
            storage.abspath(filename).startswith(
                "file://faker-file-tmp/root_tmp/rel_tmp/"
            )
        )

    def test_pathy_file_system_storage_unlink(
        self: "TestStoragesTestCase",
    ) -> None:
        """Test `PathyFileSystemStorage` `unlink`."""
        storage = PathyFileSystemStorage(
            bucket_name="faker-file-tmp",
            root_path="root_tmp",
            rel_path="rel_tmp",
        )
        with self.subTest("Test unlink by Pathy"):
            filename_1 = storage.generate_filename(prefix="", extension="tmp")
            storage.write_text(filename=filename_1, data=FAKER.text())
            self.assertTrue(storage.exists(filename_1))
            storage.unlink(filename_1)
            self.assertFalse(storage.exists(filename_1))

        with self.subTest("Test unlink by str"):
            filename_2 = storage.generate_filename(prefix="", extension="tmp")
            storage.write_text(filename=filename_2, data=FAKER.text())
            self.assertTrue(storage.exists(filename_2))
            storage.unlink(str(filename_2))
            self.assertFalse(storage.exists(filename_2))



from faker import Faker
from faker_file.providers.pdf_file import GraphicPdfFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(GraphicPdfFileProvider)  # Register provider

pdf_file = FAKER.graphic_pdf_file(size=(800, 800))



from django.contrib import admin

from .models import Upload

__all__ = ("UploadAdmin",)


@admin.register(Upload)
class UploadAdmin(admin.ModelAdmin):
    """Upload admin."""

    list_display = (
        "name",
        "description",
        "file",
    )



from typing import Any, Callable, Dict, Optional, Type, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ...base import DEFAULT_FORMAT_FUNC, BytesValue, FileMixin, StringValue
from ...constants import DEFAULT_AUDIO_MAX_NB_CHARS
from ...helpers import load_class_from_path
from ...registry import FILE_REGISTRY
from ...storages.base import BaseStorage
from ...storages.filesystem import FileSystemStorage
from ..base.mp3_generator import BaseMp3Generator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "DEFAULT_MP3_GENERATOR",
    "EDGE_TTS_MP3_GENERATOR",
    "GTTS_MP3_GENERATOR",
    "Mp3FileProvider",
)

GTTS_MP3_GENERATOR = (
    "faker_file.providers.mp3_file.generators.gtts_generator.GttsMp3Generator"
)
EDGE_TTS_MP3_GENERATOR = (
    "faker_file.providers.mp3_file.generators.edge_tts_generator"
    ".EdgeTtsMp3Generator"
)
DEFAULT_MP3_GENERATOR = GTTS_MP3_GENERATOR


class Mp3FileProvider(BaseProvider, FileMixin):
    """MP3 file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.mp3_file import Mp3FileProvider

        FAKER = Faker()
        FAKER.add_provider(Mp3FileProvider)

        file = FAKER.mp3_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.mp3_file(
            prefix="zzz",
            max_nb_chars=500,
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.mp3_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            prefix="zzz",
            max_nb_chars=500,
        )

    Default MP3 generator class is `GttsMp3Generator` which uses Google
    Text-to-Speech services to generate an MP3 file from given or
    randomly generated text. It does not require additional services to
    run and the only dependency here is the `gtts` package. You can
    however implement your own custom MP3 generator class and pass it to
    te `mp3_file` method in `mp3_generator_cls` argument instead of the
    default `GttsMp3Generator`.

    Usage with custom MP3 generator class.

    .. code-block:: python

        # Imaginary `marytts` Python library
        from marytts import MaryTTS

        # Import BaseMp3Generator
        from faker_file.providers.base.mp3_generator import (
            BaseMp3Generator,
        )

        # Define custom MP3 generator
        class MaryTtsMp3Generator(BaseMp3Generator):

            locale: str = "cmu-rms-hsmm"
            voice: str = "en_US"

            def handle_kwargs(self, **kwargs) -> None:
                # Since it's impossible to unify all TTS systems it's allowed
                # to pass arbitrary arguments to the `BaseMp3Generator`
                # constructor. Each implementation class contains its own
                # additional tuning arguments. Check the source code of the
                # implemented MP3 generators as an example.
                if "locale" in kwargs:
                    self.locale = kwargs["locale"]
                if "voice" in kwargs:
                    self.voice = kwargs["voice"]

            def generate(self) -> bytes:
                # Your implementation here. Note, that `self.content`
                # in this context is the text to make MP3 from.
                # `self.generator` would be the `Faker` or `Generator`
                # instance from which you could extract information on
                # active locale.
                # What comes below is pseudo implementation.
                mary_tts = MaryTTS(locale=self.locale, voice=self.voice)
                return mary_tts.synth_mp3(self.content)

        # Generate MP3 file from random text
        file = FAKER.mp3_file(
            mp3_generator_cls=MaryTtsMp3Generator,
        )
    """

    extension: str = "mp3"

    @overload
    def mp3_file(
        self: "Mp3FileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_AUDIO_MAX_NB_CHARS,
        content: Optional[str] = None,
        mp3_generator_cls: Optional[
            Union[str, Type[BaseMp3Generator]]
        ] = DEFAULT_MP3_GENERATOR,
        mp3_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def mp3_file(
        self: "Mp3FileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_AUDIO_MAX_NB_CHARS,
        content: Optional[str] = None,
        mp3_generator_cls: Optional[
            Union[str, Type[BaseMp3Generator]]
        ] = DEFAULT_MP3_GENERATOR,
        mp3_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def mp3_file(
        self: "Mp3FileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_AUDIO_MAX_NB_CHARS,
        content: Optional[str] = None,
        mp3_generator_cls: Optional[
            Union[str, Type[BaseMp3Generator]]
        ] = DEFAULT_MP3_GENERATOR,
        mp3_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a MP3 file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param max_nb_chars: Max number of chars for the content.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param mp3_generator_cls: Mp3 generator class.
        :param mp3_generator_kwargs: Mp3 generator kwargs.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        content = self._generate_text_content(
            max_nb_chars=max_nb_chars,
            content=content,
            format_func=format_func,
        )
        data = {"content": content, "filename": filename, "storage": storage}

        if mp3_generator_cls is None:
            mp3_generator_cls = DEFAULT_MP3_GENERATOR

        if isinstance(mp3_generator_cls, str):
            mp3_generator_cls = load_class_from_path(mp3_generator_cls)

        if not mp3_generator_kwargs:
            mp3_generator_kwargs = {}
        mp3_generator = mp3_generator_cls(
            content=content,
            generator=self.generator,
            **mp3_generator_kwargs,
        )
        _raw_content = mp3_generator.generate()

        if raw:
            raw_content = BytesValue(_raw_content)
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, _raw_content)

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from random import choice

from django.conf import settings
from factory import Faker, LazyAttribute, Trait
from factory.django import DjangoModelFactory
from faker import Faker as OriginalFaker
from faker_file.providers.bin_file import BinFileProvider
from faker_file.providers.csv_file import CsvFileProvider
from faker_file.providers.docx_file import DocxFileProvider
from faker_file.providers.eml_file import EmlFileProvider
from faker_file.providers.epub_file import EpubFileProvider
from faker_file.providers.ico_file import IcoFileProvider
from faker_file.providers.jpeg_file import JpegFileProvider
from faker_file.providers.mp3_file import Mp3FileProvider
from faker_file.providers.ods_file import OdsFileProvider
from faker_file.providers.odt_file import OdtFileProvider
from faker_file.providers.pdf_file import PdfFileProvider
from faker_file.providers.png_file import PngFileProvider
from faker_file.providers.pptx_file import PptxFileProvider
from faker_file.providers.rtf_file import RtfFileProvider
from faker_file.providers.svg_file import SvgFileProvider
from faker_file.providers.txt_file import TxtFileProvider
from faker_file.providers.webp_file import WebpFileProvider
from faker_file.providers.xlsx_file import XlsxFileProvider
from faker_file.providers.zip_file import ZipFileProvider
from faker_file.storages.filesystem import FileSystemStorage

from upload.models import Upload

FAKER = OriginalFaker()
FAKER.add_provider(BinFileProvider)
FAKER.add_provider(CsvFileProvider)
FAKER.add_provider(DocxFileProvider)
FAKER.add_provider(EmlFileProvider)
FAKER.add_provider(EpubFileProvider)
FAKER.add_provider(IcoFileProvider)
FAKER.add_provider(JpegFileProvider)
FAKER.add_provider(Mp3FileProvider)
FAKER.add_provider(OdsFileProvider)
FAKER.add_provider(OdtFileProvider)
FAKER.add_provider(PdfFileProvider)
FAKER.add_provider(PngFileProvider)
FAKER.add_provider(PptxFileProvider)
FAKER.add_provider(RtfFileProvider)
FAKER.add_provider(SvgFileProvider)
FAKER.add_provider(TxtFileProvider)
FAKER.add_provider(WebpFileProvider)
FAKER.add_provider(XlsxFileProvider)
FAKER.add_provider(ZipFileProvider)

# Define a file storage. When working with Django and FileSystemStorage
# you need to set the value of `root_path` argument to
# `settings.MEDIA_ROOT`.
STORAGE = FileSystemStorage(root_path=settings.MEDIA_ROOT, rel_path="tmp")


def random_file_generator(*args, **kwargs):
    random_provider = choice(
        [
            "bin_file",
            "csv_file",
            "docx_file",
            "eml_file",
            "epub_file",
            "ico_file",
            "jpeg_file",
            "mp3_file",
            "ods_file",
            "odt_file",
            "pdf_file",
            "png_file",
            "pptx_file",
            "rtf_file",
            "svg_file",
            "txt_file",
            "webp_file",
            "xlsx_file",
            "zip_file",
        ]
    )
    func = getattr(FAKER, random_provider)
    return func(storage=STORAGE)


class UploadFactory(DjangoModelFactory):
    """Upload factory."""

    name = Faker("text", max_nb_chars=100)
    description = Faker("text", max_nb_chars=1000)

    class Meta:
        model = Upload

    class Params:
        random_file = Trait(file=LazyAttribute(random_file_generator))


# Upload with randon file
upload = UploadFactory(random_file=True)



import os
from pathlib import Path
from random import choice
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Type,
    Union,
    overload,
)

from faker.providers import BaseProvider

from ...base import BytesValue, FileMixin, StringValue
from ...helpers import load_class_from_path
from ...storages.base import BaseStorage
from ...storages.filesystem import FileSystemStorage
from ..base.text_augmenter import BaseTextAugmenter
from ..base.text_extractor import BaseTextExtractor
from ..helpers.inner import (
    create_inner_docx_file,
    create_inner_eml_file,
    create_inner_epub_file,
    create_inner_odt_file,
    create_inner_pdf_file,
    create_inner_rtf_file,
    create_inner_txt_file,
)

# from .augmenters.nlpaug_augmenter import ContextualWordEmbeddingsAugmenter
# from .extractors.tika_extractor import TikaTextExtractor

# Full list:
# create_inner_bin_file: N/A
# create_inner_bmp_file: Not supported
# create_inner_csv_file: Not supported
# create_inner_docx_file: Supported
# create_inner_eml_file: Supported
# create_inner_epub_file: Supported
# create_inner_generic_file: Not supported
# create_inner_gif_file: Not supported
# create_inner_ico_file: Not supported
# create_inner_jpeg_file: Not supported
# create_inner_json_file: Not supported
# create_inner_mp3_file: Not supported
# create_inner_ods_file: Not supported
# create_inner_odt_file: Supported
# create_inner_pdf_file: Supported
# create_inner_png_file: Not supported
# create_inner_pptx_file: Not supported
# create_inner_rtf_file: Supported
# create_inner_svg_file: Not supported
# create_inner_tar_file: Not supported
# create_inner_txt_file: Supported
# create_inner_webp_file: Not supported
# create_inner_xlsx_file: Not supported
# create_inner_xml_file: Not supported
# create_inner_zip_file: Not supported


__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "AugmentFileFromDirProvider",
    "NLPAUG_AUGMENTER",
    "TEXTAUGMENT_AUGMENTER",
    "DEFAULT_AUGMENTER",
    "TIKA_EXTRACTOR",
    "DEFAULT_EXTRACTOR",
    "FILE_TYPE_TO_INNER_FUNC_MAPPING",
    "EXTENSIONS",
)

NLPAUG_AUGMENTER = (
    "faker_file.providers.augment_file_from_dir.augmenters.nlpaug_augmenter"
    ".ContextualWordEmbeddingsAugmenter"
)
TEXTAUGMENT_AUGMENTER = (
    "faker_file.providers.augment_file_from_dir.augmenters"
    ".textaugment_augmenter.EDATextaugmentAugmenter"
)
DEFAULT_AUGMENTER = TEXTAUGMENT_AUGMENTER

TIKA_EXTRACTOR = (
    "faker_file.providers.augment_file_from_dir.extractors.tika_extractor"
    ".TikaTextExtractor"
)
DEFAULT_EXTRACTOR = TIKA_EXTRACTOR

FILE_TYPE_TO_INNER_FUNC_MAPPING: Dict[str, Callable] = {
    "docx": create_inner_docx_file,
    "eml": create_inner_eml_file,
    "epub": create_inner_epub_file,
    "odt": create_inner_odt_file,
    "pdf": create_inner_pdf_file,
    "rtf": create_inner_rtf_file,
    "txt": create_inner_txt_file,
}

EXTENSIONS = set(FILE_TYPE_TO_INNER_FUNC_MAPPING.keys())


class AugmentFileFromDirProvider(BaseProvider, FileMixin):
    """Augment file from given directory provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.augment_file_from_dir import (
            AugmentFileFromDirProvider,
        )

        FAKER = Faker()
        FAKER.add_provider(AugmentFileFromDirProvider)

        file = FAKER.augment_file_from_dir(
            source_dir_path="/tmp/tmp/",
        )

    Usage example with options:

        .. code-block:: python

        file = FAKER.augment_file_from_dir(
            source_dir_path="/tmp/tmp/",
            prefix="zzz",
            extensions={"docx", "pdf"}
        )
    """

    extension: str = ""

    @overload
    def augment_file_from_dir(
        self: "AugmentFileFromDirProvider",
        source_dir_path: str,
        extensions: Optional[Iterable[str]] = None,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        wrap_chars_after: Optional[int] = None,
        text_extractor_cls: Optional[
            Union[str, Type[BaseTextExtractor]]
        ] = DEFAULT_EXTRACTOR,
        text_extractor_kwargs: Optional[Dict[str, Any]] = None,
        text_augmenter_cls: Optional[
            Union[str, Type[BaseTextAugmenter]]
        ] = DEFAULT_AUGMENTER,
        text_augmenter_kwargs: Optional[Dict[str, Any]] = None,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def augment_file_from_dir(
        self: "AugmentFileFromDirProvider",
        source_dir_path: str,
        extensions: Optional[Iterable[str]] = None,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        wrap_chars_after: Optional[int] = None,
        text_extractor_cls: Optional[
            Union[str, Type[BaseTextExtractor]]
        ] = DEFAULT_EXTRACTOR,
        text_extractor_kwargs: Optional[Dict[str, Any]] = None,
        text_augmenter_cls: Optional[
            Union[str, Type[BaseTextAugmenter]]
        ] = DEFAULT_AUGMENTER,
        text_augmenter_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> StringValue: ...

    def augment_file_from_dir(
        self: "AugmentFileFromDirProvider",
        source_dir_path: str,
        extensions: Optional[Iterable[str]] = None,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        wrap_chars_after: Optional[int] = None,
        text_extractor_cls: Optional[
            Union[str, Type[BaseTextExtractor]]
        ] = DEFAULT_EXTRACTOR,
        text_extractor_kwargs: Optional[Dict[str, Any]] = None,
        text_augmenter_cls: Optional[
            Union[str, Type[BaseTextAugmenter]]
        ] = DEFAULT_AUGMENTER,
        text_augmenter_kwargs: Optional[Dict[str, Any]] = None,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Augment a random file from given directory.

        :param source_dir_path: Source files directory.
        :param extensions: Allowed extensions.
        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param text_extractor_cls: Text extractor class.
        :param text_extractor_kwargs: Text extractor kwargs.
        :param text_augmenter_cls: Text augmenter class.
        :param text_augmenter_kwargs: Text augmenter kwargs.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        if extensions is None:
            extensions = EXTENSIONS

        # Specific
        source_file_choices = [
            os.path.join(source_dir_path, _f)
            for _f in os.listdir(source_dir_path)
            if (
                os.path.isfile(os.path.join(source_dir_path, _f))
                and os.path.splitext(_f)[1][1:] in extensions
            )
        ]
        source_file_path = choice(source_file_choices)
        source_file = Path(source_file_path)

        if text_extractor_cls is None:
            text_extractor_cls = DEFAULT_EXTRACTOR

        if isinstance(text_extractor_cls, str):
            text_extractor_cls = load_class_from_path(text_extractor_cls)

        if not text_extractor_kwargs:
            text_extractor_kwargs = {}

        text_extractor = text_extractor_cls(**text_extractor_kwargs)
        extracted_content = text_extractor.extract(source_file)
        file_type = source_file.suffix[1:]

        if text_augmenter_cls is None:
            text_augmenter_cls = DEFAULT_AUGMENTER

        if isinstance(text_augmenter_cls, str):
            text_augmenter_cls = load_class_from_path(text_augmenter_cls)

        if not text_augmenter_kwargs:
            text_augmenter_kwargs = {}

        text_augmenter = text_augmenter_cls(**text_augmenter_kwargs)
        content = text_augmenter.augment(extracted_content)

        return FILE_TYPE_TO_INNER_FUNC_MAPPING[file_type](
            storage=storage,
            basename=basename,
            prefix=prefix,
            content=content,
            wrap_chars_after=wrap_chars_after,
            raw=raw,
        )



import logging
from typing import Any, Dict, Union

import pdfkit
from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider

from ....base import DynamicTemplate, StringList
from ....constants import DEFAULT_FILE_ENCODING
from ...base.pdf_generator import BasePdfGenerator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("PdfkitPdfGenerator",)

LOGGER = logging.getLogger(__name__)


class PdfkitPdfGenerator(BasePdfGenerator):
    """Pdfkit PDF generator.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.pdf_file import PdfFileProvider
        from faker_file.providers.pdf_file.generators.pdfkit_generator import (
            PdfkitPdfGenerator
        )

        FAKER = Faker()
        FAKER.add_provider(PdfFileProvider)

        file = FAKER.pdf_file(pdf_generator_cls=PdfkitPdfGenerator)

    Using `DynamicTemplate`:

    .. code-block:: python

        from faker_file.base import DynamicTemplate
        from faker_file.contrib.pdf_file.pdfkit_snippets import (
            add_h1_heading,
            add_h2_heading,
            add_h3_heading,
            add_h4_heading,
            add_h5_heading,
            add_h6_heading,
            add_heading,
            add_page_break,
            add_paragraph,
            add_picture,
            add_table,
        )

        # Create a file with lots of elements
        file = FAKER.pdf_file(
            pdf_generator_cls=PdfkitPdfGenerator,
            content=DynamicTemplate(
                [
                    (add_h1_heading, {}),
                    (add_paragraph, {}),
                    (add_h2_heading, {}),
                    (add_h3_heading, {}),
                    (add_h4_heading, {}),
                    (add_h5_heading, {}),
                    (add_h6_heading, {}),
                    (add_paragraph, {}),
                    (add_picture, {}),
                    (add_page_break, {}),
                    (add_h6_heading, {}),
                    (add_table, {}),
                    (add_paragraph, {}),
                ]
            )
        )
    """

    encoding: str = DEFAULT_FILE_ENCODING

    def handle_kwargs(self: "PdfkitPdfGenerator", **kwargs) -> None:
        """Handle kwargs."""
        if "encoding" in kwargs:
            self.encoding = kwargs["encoding"]

    def generate(
        self: "PdfkitPdfGenerator",
        content: Union[str, DynamicTemplate],
        data: Dict[str, Any],
        provider: Union[Faker, Generator, Provider],
        **kwargs,
    ) -> bytes:
        """Generate PDF."""
        options = {"quiet": ""}
        if self.encoding is not None:
            options["encoding"] = self.encoding

        if isinstance(content, DynamicTemplate):
            _content = StringList()
            for counter, (ct_modifier, ct_modifier_kwargs) in enumerate(
                content.content_modifiers
            ):
                ct_modifier(
                    provider,
                    self,
                    _content,
                    data,
                    counter,
                    **ct_modifier_kwargs,
                )
        else:
            _content = f"<pre style='white-space: pre-wrap;'>{content}</pre>"

        return pdfkit.from_string(
            str(_content),
            options=options,
        )



# Required imports
from faker import Faker
from faker_file.providers.pdf_file import PdfFileProvider
from faker_file.providers.pdf_file.generators.pdfkit_generator import (
    PdfkitPdfGenerator,
)

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(PdfFileProvider)  # Register PdfFileProvider

# Generate PDF file using `pdfkit`
pdf_file = FAKER.pdf_file(pdf_generator_cls=PdfkitPdfGenerator)



from faker import Faker
from faker_file.providers.augment_image_from_path import (
    AugmentImageFromPathProvider,
)
from faker_file.providers.augment_random_image_from_dir import (
    AugmentRandomImageFromDirProvider,
)
from faker_file.providers.image.augment import (
    add_brightness,
    decrease_contrast,
    flip_horizontal,
    flip_vertical,
    resize_height,
    resize_width,
)
from faker_file.providers.png_file import GraphicPngFileProvider

FAKER = Faker()
FAKER.add_provider(AugmentImageFromPathProvider)
FAKER.add_provider(AugmentRandomImageFromDirProvider)
FAKER.add_provider(GraphicPngFileProvider)

# Create a couple of graphic images to augment later on.
FAKER.graphic_png_file(basename="01")  # One named 01.png
# And 5 more with random names.
for __ in range(5):
    FAKER.graphic_png_file()

# We assumed that directory "/tmp/tmp/" exists and contains
# image files, among which "01.png". Augmentations will be applied
# sequentially, one by one until all fulfilled. If you wish to apply only
# a random number of augmentations, but not all, pass the `num_steps`
# argument, with value less than the number of `augmentations` provided.
augmented_image_file = FAKER.augment_image_from_path(
    path="/tmp/tmp/01.png",
    augmentations=[
        (flip_horizontal, {}),
        (flip_vertical, {}),
        (decrease_contrast, {}),
        (add_brightness, {}),
        (resize_width, {"lower": 0.9, "upper": 1.1}),
        (resize_height, {"lower": 0.9, "upper": 1.1}),
    ],
    prefix="augmented_image_01_",
    # num_steps=3,
)

augmented_random_image_file = FAKER.augment_random_image_from_dir(
    source_dir_path="/tmp/tmp/",
    augmentations=[
        (flip_horizontal, {}),
        (flip_vertical, {}),
        (decrease_contrast, {}),
        (add_brightness, {}),
        (resize_width, {"lower": 0.9, "upper": 1.1}),
        (resize_height, {"lower": 0.9, "upper": 1.1}),
    ],
    prefix="augmented_random_image_",
    # num_steps=3,
)



from typing import Callable, Optional, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import DEFAULT_FORMAT_FUNC, BytesValue, FileMixin, StringValue
from ..constants import DEFAULT_TEXT_MAX_NB_CHARS
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("RtfFileProvider",)


class RtfFileProvider(BaseProvider, FileMixin):
    """RTF file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.rtf_file import RtfFileProvider

        FAKER = Faker()
        FAKER.add_provider(RtfFileProvider)

        file = FAKER.rtf_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.rtf_file(
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.rtf_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )
    """

    extension: str = "rtf"

    @overload
    def rtf_file(
        self: "RtfFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def rtf_file(
        self: "RtfFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def rtf_file(
        self: "RtfFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a RTF file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param max_nb_chars: Max number of chars for the content.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        content = self._generate_text_content(
            max_nb_chars=max_nb_chars,
            wrap_chars_after=wrap_chars_after,
            content=content,
            format_func=format_func,
        )

        data = {"content": content, "filename": filename, "storage": storage}

        if raw:
            raw_content = BytesValue(content.encode("utf8"))
            raw_content.data = data
            return raw_content

        storage.write_text(filename, content)

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from . import db

__all__ = ("Upload",)


class Upload(db.Model):
    """Upload model."""

    __tablename__ = "upload"
    __table_args__ = {"extend_existing": True}

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(64), unique=True)
    description = db.Column(db.Integer(), nullable=True)
    file = db.Column(db.Unicode(255))

    def __str__(self) -> str:
        return f"{self.name}"
